# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""StandardCheckpointHandler class."""

from __future__ import annotations

import dataclasses
import numbers
from typing import Any, List, Optional

from absl import logging
from etils import epath
import jax
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint import checkpoint_utils
from orbax.checkpoint import options as options_lib
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.handlers import async_checkpoint_handler
from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
from orbax.checkpoint._src.metadata import pytree_metadata_options as pytree_metadata_options_lib
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.tree import types as tree_types
from orbax.checkpoint._src.tree import utils as tree_utils


PyTree = Any
CheckpointArgs = checkpoint_args.CheckpointArgs
PyTreeMetadataOptions = pytree_metadata_options_lib.PyTreeMetadataOptions
register_with_handler = checkpoint_args.register_with_handler


class StandardCheckpointHandler(
    async_checkpoint_handler.AsyncCheckpointHandler
):
  """A CheckpointHandler implementation for any PyTree structure.

  See JAX documentation for more information on what constitutes a "PyTree".
  This handler is capable of saving and restoring PyTrees with leaves of type
  Python scalar, np.ndarray, and jax.Array

  As with all `CheckpointHandler` subclasses, `StandardCheckpointHandler` should
  only be used in conjunction with a `Checkpointer` (or subclass). By itself,
  the `CheckpointHandler` is non-atomic.

  Example::

    ckptr = Checkpointer(StandardCheckpointHandler())
    # OR
    ckptr = StandardCheckpointer()

  If you find that your use case is not covered by `StandardCheckpointHandler`,
  consider using the parent class directly, or explore a custom implementation
  of `CheckpointHandler`.
  """

  def __init__(
      self,
      *,
      save_concurrent_gb: int = 96,
      restore_concurrent_gb: int = 96,
      multiprocessing_options: options_lib.MultiprocessingOptions = (
          options_lib.MultiprocessingOptions()
      ),
      pytree_metadata_options: PyTreeMetadataOptions = (
          pytree_metadata_options_lib.PYTREE_METADATA_OPTIONS
      ),
  ):
    """Creates StandardCheckpointHandler.

    Args:
      save_concurrent_gb: max concurrent GB that are allowed to be saved. Can
        help to reduce the possibility of OOM's when large checkpoints are
        saved.
      restore_concurrent_gb: max concurrent GB that are allowed to be restored.
        Can help to reduce the possibility of OOM's when large checkpoints are
        restored.
      multiprocessing_options: See orbax.checkpoint.options.
      pytree_metadata_options: Options to control types like tuple and
        namedtuple in pytree metadata.
    """
    self._supported_types = checkpoint_utils.STANDARD_ARRAY_TYPES
    self._impl = pytree_checkpoint_handler.PyTreeCheckpointHandler(
        save_concurrent_gb=save_concurrent_gb,
        restore_concurrent_gb=restore_concurrent_gb,
        multiprocessing_options=multiprocessing_options,
        pytree_metadata_options=pytree_metadata_options,
    )

  def _validate_save_state(
      self, item: PyTree, save_args: Optional[PyTree] = None
  ):
    if item is None:
      raise ValueError('Must provide item to save.')
    if isinstance(item, jax.Array | numbers.Number):
      raise ValueError(
          'StandardCheckpointHandler / StandardSave does not support single '
          'arrays or scalars. Use ArrayCheckpointHandler / ArraySave'
      )
    if save_args is None:
      save_args = jax.tree.map(lambda x: None, item)

    def _check_input(k, x, arg):
      if arg is not None:
        if arg.aggregate:
          raise ValueError(f'Unsupported option `aggregate` for key: {k}.')
      if not isinstance(x, self._supported_types):
        k = tree_utils.tuple_path_from_keypath(k)
        raise ValueError(f'Unsupported type: {type(x)} for key: {k}.')

    jax.tree_util.tree_map_with_path(_check_input, item, save_args)

  def _validate_restore_state(self, item: PyTree):
    def _check_input(k, x):
      if not isinstance(x, self._supported_types) and not isinstance(
          x, jax.ShapeDtypeStruct
      ):
        k = tree_utils.tuple_path_from_keypath(k)
        raise ValueError(f'Unsupported type: {type(x)} for key: {k}.')

    jax.tree_util.tree_map_with_path(_check_input, item)

  async def async_save(
      self,
      directory: epath.Path,
      item: Optional[PyTree] = None,
      save_args: Optional[PyTree] = None,
      args: Optional[StandardSaveArgs] = None,
  ) -> Optional[List[future.Future]]:
    """Saves a PyTree of array-like objects. See PyTreeCheckpointHandler."""
    if isinstance(item, CheckpointArgs):
      raise ValueError(
          'Make sure to specify kwarg name `args=` when providing'
          ' `StandardSaveArgs`.'
      )
    custom_metadata = None
    if args is not None:
      item = args.item
      save_args = args.save_args
      custom_metadata = args.custom_metadata

    self._validate_save_state(item, save_args=save_args)
    return await self._impl.async_save(
        directory,
        args=pytree_checkpoint_handler.PyTreeSaveArgs(
            item=item,
            save_args=save_args,
            custom_metadata=custom_metadata,
        ),
    )

  def save(self, directory: epath.Path, *args, **kwargs):
    """Saves the provided item synchronously."""

    async def async_save(*args, **kwargs):
      commit_futures = await self.async_save(*args, **kwargs)  # pytype: disable=bad-return-type
      # Futures are already running, so sequential waiting is equivalent to
      # concurrent waiting.
      if commit_futures:  # May be None.
        for f in commit_futures:
          f.result()  # Block on result.

    asyncio_utils.run_sync(async_save(directory, *args, **kwargs))

  def restore(
      self,
      directory: epath.Path,
      item: Optional[PyTree] = None,
      args: Optional[StandardRestoreArgs] = None,
  ) -> PyTree:
    """Restores a PyTree. See PyTreeCheckpointHandler.

    Example::

      ckptr = StandardCheckpointer()
      item = {
          'layer0': {
              'w': jax.Array(...),
              'b': np.ndarray(...),
          },
      }
      ckptr.save(dir, StandardSaveArgs(item))

      target = {
          'layer0': {
              'w': jax.ShapeDtypeStruct(...),
              'b': jax.Array(...),
          },
      }
      ckptr.restore(dir, StandardRestoreArgs(target))

    Args:
      directory: path from which to restore.
      item: Deprecated, use `args`.
      args: `StandardRestoreArgs` (see below).

    Returns:
      a restored PyTree.
    """
    if isinstance(item, CheckpointArgs):
      raise ValueError(
          'Make sure to specify kwarg name `args=` when providing'
          ' `StandardRestoreArgs`.'
      )
    if not args:
      args = StandardRestoreArgs(item=item)
    if args.item is not None:
      self._validate_restore_state(args.item)
      restore_args = checkpoint_utils.construct_restore_args(
          args.item, support_layout=args.support_layout
      )
    else:
      logging.warning(
          '`StandardCheckpointHandler` expects a target tree to be provided for'
          ' restore. Not doing so is generally UNSAFE unless you know the'
          ' present topology to be the same one as the checkpoint was saved'
          ' under.'
      )
      restore_args = checkpoint_utils.construct_restore_args(
          self.metadata(directory)
      )

    def _replace_strict(
        arg: pytree_checkpoint_handler.RestoreArgs,
    ) -> pytree_checkpoint_handler.RestoreArgs:
      if hasattr(arg, 'strict'):
        return dataclasses.replace(arg, strict=False)
      return arg

    if not args.strict:
      restore_args = jax.tree.map(_replace_strict, restore_args)
    return self._impl.restore(
        directory,
        args=pytree_checkpoint_handler.PyTreeRestoreArgs(
            item=args.item, restore_args=restore_args
        ),
    )

  def metadata(self, directory: epath.Path) -> tree_metadata.TreeMetadata:
    """Returns metadata about the saved item."""
    return self._impl.metadata(directory)

  def finalize(self, directory: epath.Path) -> None:
    self._impl.finalize(directory)

  def close(self):
    self._impl.close()


@register_with_handler(StandardCheckpointHandler, for_save=True)
@dataclasses.dataclass
class StandardSaveArgs(CheckpointArgs):
  """Parameters for saving a standard PyTree.

  Also see `PyTreeSave` for additional options.

  Attributes:
    item (required): a PyTree to be saved.
    save_args: a PyTree with the same structure of `item`, which consists of
      `ocp.SaveArgs` objects as values. `None` can be used for values where no
      `SaveArgs` are specified.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """

  item: PyTree
  save_args: Optional[PyTree] = None
  custom_metadata: tree_types.JsonType | None = None

  def __post_init__(self):
    if isinstance(self.item, tree_metadata.TreeMetadata):
      raise ValueError('Cannot save TreeMetadata.')


@register_with_handler(StandardCheckpointHandler, for_restore=True)
@dataclasses.dataclass
class StandardRestoreArgs(CheckpointArgs):
  """Parameters for restoring a standard PyTree.

  Also see `PyTreeRestore` for additional options.

  Attributes (all optional):
    item: target PyTree. Currently non-optional. Values may be either real
        array or scalar values, or they may be jax.ShapeDtypeStruct, or
        `ocp.metadata.value.Metadata` objects (which come from calling the
        `metadata` method). If real values are provided,
        that value will be restored as the given type, with
        the given properties. If jax.ShapeDtypeStruct is provided, the value
        will be restored as np.ndarray, unless `sharding` is specified. If
        `item` is a custom PyTree class, the tree will be restored with the
        same structure as provided. If not provided, restores as a serialized
        nested dict representation of the custom class.
        `TreeMetadata` is also allowed as the tree used to
        define the restored structure.
    strict: if False, restoration allows silent truncating/padding of arrays if
        the stored array shape does not match the target shape. Otherwise,
        raises an error.
    support_layout: if True, restores with the layouts in `item`.
  """

  item: Optional[PyTree] = None
  strict: bool = True
  support_layout: bool = False

  def __post_init__(self):
    if isinstance(self.item, tree_metadata.TreeMetadata):
      self.item = self.item.tree



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BasePyTreeCheckpointHandler class.

Implementation of `CheckpointHandler` interface dealing with JAX PyTrees. Much
of the underlying reading/writing logic for individual leaf types can be
customized, and is delegated to the `TypeHandler` class.
"""

from __future__ import annotations

import asyncio
import dataclasses
import functools
import json
import sys
import time
from typing import Any, List, Optional, Sequence, Tuple, Union
import uuid

from absl import logging
from etils import epath
import humanize
import jax
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint import options as options_lib
from orbax.checkpoint import utils
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.futures import synchronization
from orbax.checkpoint._src.handlers import async_checkpoint_handler
from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import format_utils
from orbax.checkpoint._src.serialization import serialization
from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint._src.serialization import types
from orbax.checkpoint._src.tree import types as tree_types
from orbax.checkpoint._src.tree import utils as tree_utils
import tensorstore as ts



PyTree = Any
TupleKey = Tuple[str, ...]
RestoreArgs = type_handlers.RestoreArgs
ArrayRestoreArgs = type_handlers.ArrayRestoreArgs
SaveArgs = type_handlers.SaveArgs
ParamInfo = type_handlers.ParamInfo
TypeHandler = type_handlers.TypeHandler
TypeHandlerRegistry = type_handlers.TypeHandlerRegistry

# TODO(b/298487158) Clean up protected access.
LimitInFlightBytes = type_handlers.LimitInFlightBytes
CheckpointArgs = checkpoint_args.CheckpointArgs
register_with_handler = checkpoint_args.register_with_handler
get_param_names = tree_utils.get_param_names
PYTREE_METADATA_FILE = format_utils.PYTREE_METADATA_FILE

DEFAULT_CONCURRENT_GB = 96



def _default_sizeof_values(values: Sequence[Any]) -> Sequence[int]:
  return [sys.getsizeof(v) for v in values]


def _get_batch_memory_size(
    handler: TypeHandler, values: Sequence[Any]
) -> Tuple[int, int]:
  """Gets memory size for a batch of leaf values."""
  try:
    write_sizes, read_sizes = zip(*handler.memory_size(values))
  except NotImplementedError:
    logging.warning(
        '`memory_size` is not implemented for `TypeHandler` of type: %s. Using'
        ' the a default implementation to measure value memory consumption that'
        ' may result in inaccurate estimation.',
        type(handler),
    )
    write_sizes = read_sizes = _default_sizeof_values(values)
  assert len(write_sizes) == len(values)
  assert len(read_sizes) == len(values)
  return sum(write_sizes), sum(read_sizes)


def _log_io_metrics(
    size: int,
    start_time: float,
    bytes_per_sec_metric: str,
    bytes_metric: Optional[str] = None,
):
  """Logs the bytes per second metric."""
  time_elapsed = time.time() - start_time
  bytes_per_sec = (
      float('nan') if time_elapsed == 0 else float(size) / time_elapsed
  )
  logging.info(
      '[process=%d] %s: %s/s (total bytes: %s) (time elapsed: %s) (per-host)',
      multihost.process_index(),
      bytes_per_sec_metric,
      humanize.naturalsize(bytes_per_sec, binary=True),
      humanize.naturalsize(size, binary=True),
      humanize.naturaldelta(time_elapsed, minimum_unit='microseconds'),
  )
  jax.monitoring.record_event_duration_secs(bytes_per_sec_metric, bytes_per_sec)
  if bytes_metric is not None:
    jax.monitoring.record_event(bytes_metric, bytes=size)


@dataclasses.dataclass
class _BatchRequest:
  """Represents a a request for batched serialization or deserialization.

  Attributes:
    handler: Used to serialize or deserialize the parameters.
    keys: Used to identify the original tree keys so that the PyTree can be
      reconstructed.
    values: Values to serialize.
    infos: ParamInfos.
    args: List of SaveArgs or RestoreArgs.
  """

  handler: TypeHandler
  keys: List[str]
  values: List[Any]
  infos: List[ParamInfo]
  args: List[Union[SaveArgs, RestoreArgs]]

  def __post_init__(self):
    length = len(self.values)
    if not all((
        length == len(self.infos),
        length == len(self.args),
        length == len(self.keys),
    )):
      raise AssertionError('Found `_BatchRequest` with mismatched parameters.')


def batched_serialization_requests(
    tree: PyTree,
    param_infos: PyTree,
    args: PyTree,
    registry: TypeHandlerRegistry,
) -> List[_BatchRequest]:
  """Gets a list of batched serialization or deserialization requests."""
  grouped = {}

  def _group_value(
      keypath: Tuple[Any, ...],
      info: ParamInfo,
      value: Union[Any, tree_metadata.ValueMetadataEntry],
      arg: Union[SaveArgs, RestoreArgs],
  ):
    nonlocal grouped
    tuple_key = tree_utils.tuple_path_from_keypath(keypath)
    if info.skip_deserialize:
      return

    if isinstance(arg, RestoreArgs):
      assert isinstance(value, tree_metadata.ValueMetadataEntry), type(value)
      metadata_restore_type = value.value_type
      requested_restore_type = arg.restore_type or metadata_restore_type
      # TODO(cpgaffney): Add a warning message if the requested_restore_type
      # is not the same as the metadata_restore_type.
      if empty_values.is_empty_typestr(requested_restore_type):
        # Skip deserialization of empty node using TypeHandler.
        return
      type_for_registry_lookup = requested_restore_type
    elif isinstance(arg, SaveArgs):
      # Skip serialization of empty node using TypeHandler.
      if tree_utils.is_empty_node(value):
        return
      type_for_registry_lookup = type(value)
    else:
      raise AssertionError(
          f'Expected `RestoreArgs` or `SaveArgs`. Got {type(arg)}.'
      )

    try:
      handler = registry.get(type_for_registry_lookup)
    except ValueError as e:
      raise ValueError(
          f'TypeHandler lookup failed for: type={type_for_registry_lookup},'
          f' keypath={keypath}, ParamInfo={info}, RestoreArgs={arg},'
          f' value={value}'
      ) from e

    if handler not in grouped:
      grouped[handler] = _BatchRequest(handler, [], [], [], [])
    request = grouped[handler]
    grouped[handler] = dataclasses.replace(
        request,
        keys=request.keys + [tuple_key],
        values=request.values + [value],
        infos=request.infos + [info],
        args=request.args + [arg],
    )

  jax.tree_util.tree_map_with_path(
      _group_value,
      param_infos,
      tree,
      args,
  )
  return list(grouped.values())


def _fill_missing_save_or_restore_args(
    item: PyTree, args: Optional[PyTree], *, mode: str
) -> PyTree:
  """Fills in missing values in the tree of SaveArgs or RestoreArgs.

  Values may be "missing" because of empty nodes in `item`. After returning, all
  keys in `item`, with empty nodes or not, will have a corresponding value
  in the result.

  Args:
    item: tree to save or target to restore.
    args: tree of SaveArgs or RestoreArgs. May be None, if the user did not
      provide it.
    mode: 'save' or 'restore'.

  Returns:
    A tree of SaveArgs or RestoreArgs with missing values filled in.
  """

  # Because of empty states, the user-provided args may not contain
  # all necessary arguments. These should be filled in with default args.
  def _maybe_set_default_save_args(_, leaf_args):
    if isinstance(leaf_args, (SaveArgs, RestoreArgs)):
      return leaf_args
    elif mode == 'save':
      return SaveArgs()
    elif mode == 'restore':
      return RestoreArgs()
    else:
      raise ValueError(f'Unknown mode: {mode}.')

  return jax.tree_util.tree_map(
      _maybe_set_default_save_args,
      item,
      item if args is None else args,
      is_leaf=utils.is_empty_or_leaf,
  )




class BasePyTreeCheckpointHandler(
    async_checkpoint_handler.AsyncCheckpointHandler
):
  """A CheckpointHandler implementation for any PyTree structure.

  Largely serves as the implementation for `PyTreeCheckpointHandler`. Users are
  advised not to use this class directly.
  """

  def __init__(
      self,
      *,
      save_concurrent_bytes: Optional[int] = None,
      restore_concurrent_bytes: Optional[int] = None,
      use_ocdbt: bool = True,
      use_zarr3: bool = False,
      multiprocessing_options: options_lib.MultiprocessingOptions = options_lib.MultiprocessingOptions(),
      type_handler_registry: TypeHandlerRegistry = type_handlers.GLOBAL_TYPE_HANDLER_REGISTRY,
      enable_post_merge_validation: bool = True,
      pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
          tree_metadata.PYTREE_METADATA_OPTIONS
      ),
      array_metadata_validator: array_metadata_store_lib.Validator = (
          array_metadata_store_lib.Validator()
      ),
  ):
    """Creates BasePyTreeCheckpointHandler.

    Args:
      save_concurrent_bytes: max concurrent bytes that are allowed to be
        written. Can help to reduce the possibility of OOM's when large
        checkpoints are saved.
      restore_concurrent_bytes: max concurrent bytes that are allowed to be
        restored. Can help to reduce the possibility of OOM's when large
        checkpoints are restored.
      use_ocdbt: Whether to use OCDBT format for saving.
      use_zarr3: If True, use Zarr ver3 otherwise Zarr ver2.
      multiprocessing_options: See orbax.checkpoint.options.
      type_handler_registry: a type_handlers.TypeHandlerRegistry. If not
        specified, the global type handler registry will be used. # BEGIN
      enable_descriptor: If True, logs a Descriptor proto that contains lineage
      enable_post_merge_validation: If True, enables validation of the
        parameters after the finalize step.
      pytree_metadata_options: `PyTreeMetadataOptions` to manage metadata.
      array_metadata_validator: Validator for ArrayMetadata.
    """
    self._save_concurrent_bytes = save_concurrent_bytes
    self._restore_concurrent_bytes = restore_concurrent_bytes
    self._use_ocdbt = use_ocdbt
    self._use_zarr3 = use_zarr3
    self._primary_host = multiprocessing_options.primary_host
    self._type_handler_registry = type_handler_registry
    self._enable_post_merge_validation = enable_post_merge_validation
    self._pytree_metadata_options = pytree_metadata_options
    # Get ArrayMetadata Store from TypeHandler for jax.Array.
    # ArrayMetadata persistence is only supported for jax.Array.
    self._array_metadata_store = (
        array_metadata_store_lib.resolve_array_metadata_store(
            type_handler_registry
        )
    )
    if self._array_metadata_store:
      self._array_metadata_store.set_primary_host(self._primary_host)
    self._array_metadata_validator = array_metadata_validator


    jax.monitoring.record_event(
        '/jax/orbax/pytree_checkpoint_handler/init/ocdbt'
    )
    logging.info(
        'Created BasePyTreeCheckpointHandler: pytree_metadata_options=%s,'
        ' array_metadata_store=%s',
        self._pytree_metadata_options,
        self._array_metadata_store,
    )

  def get_param_names(self, item: PyTree) -> PyTree:
    """Gets parameter names for PyTree elements."""
    return get_param_names(item)

  def _get_param_infos(
      self,
      item: PyTree,
      directory: epath.Path,
      *,
      use_ocdbt: bool = True,
      use_zarr3: Optional[bool] = None,
      ocdbt_target_data_file_size: Optional[int] = None,
      enable_pinned_host_transfer: bool = False,
      byte_limiter: Optional[serialization.ByteLimiter] = None,
      raise_array_data_missing_error: bool = True,
  ) -> PyTree:
    """Returns parameter information for elements in `item`.

    At minimum, this method should extract the names of each parameter for
    saving/restoring.

    Args:
      item: a PyTree to extract information from.
      directory: a directory where checkpoint files are located.
      use_ocdbt: Whether to use OCDBT for writing or reading.
      use_zarr3: Whether to use zarr3.
      ocdbt_target_data_file_size: Specifies the target size (in bytes) of each
        OCDBT data file.
      enable_pinned_host_transfer: See ParamInfo docs.
      byte_limiter: ByteLimiter object.
      raise_array_data_missing_error: See documentation in ParamInfo.

    Returns:
      A PyTree matching `item` of ParamInfo.
    """
    if use_zarr3 is None:
      use_zarr3 = self._use_zarr3
    names = self.get_param_names(item)
    ts_context = ts_utils.get_ts_context(use_ocdbt=use_ocdbt)

    def _param_info(name, value):
      skip_deserialize = False
      if isinstance(value, tree_metadata.ValueMetadataEntry):
        skip_deserialize = value.skip_deserialize
      return ParamInfo(
          name=name,
          path=(directory / name),
          parent_dir=directory,
          skip_deserialize=skip_deserialize,
          is_ocdbt_checkpoint=use_ocdbt,
          use_zarr3=use_zarr3,
          enable_pinned_host_transfer=enable_pinned_host_transfer,
          ocdbt_target_data_file_size=ocdbt_target_data_file_size,
          byte_limiter=byte_limiter,
          ts_context=ts_context,
          value_typestr=types.get_param_typestr(
              value, self._type_handler_registry, self._pytree_metadata_options
          ),
          raise_array_data_missing_error=raise_array_data_missing_error,
      )

    return jax.tree.map(
        _param_info, names, item, is_leaf=utils.is_empty_or_leaf
    )

  async def async_save(
      self,
      directory: epath.Path,
      args: BasePyTreeSaveArgs,
  ) -> Optional[List[future.Future]]:
    """Saves a PyTree to a given directory.

    This operation is compatible with a multi-host, multi-device setting. Tree
    leaf values must be supported by the type_handler_registry given in the
    constructor. Standard supported types include Python scalars, `np.ndarray`,
    `jax.Array`, and strings.

    After saving, all files will be located in "directory/".
    A JSON metadata file will be present to store the tree structure.

    Example usage::

      ckptr = Checkpointer(BasePyTreeCheckpointHandler())
      item = {
          'layer0': {
              'w': np.ndarray(...),
              'b': np.ndarray(...),
          },
          'layer1': {
              'w': np.ndarray(...),
              'b': np.ndarray(...),
          },
      }
      # Note: save_args may be None if no customization is desired for saved
      # parameters.
      # Eventually calls through to `async_save`.
      ckptr.save(path, item, save_args)

    Args:
      directory: save location directory.
      args: `BasePyTreeSaveArgs` (see below).

    Returns:
      A Future that will commit the data to `directory` when awaited. Copying
      the data from its source will be awaited in this function.
    """
    start_time = time.time()
    item = args.item
    if not item:
      raise ValueError('Found empty item.')
    save_args = args.save_args
    ocdbt_target_data_file_size = args.ocdbt_target_data_file_size
    custom_metadata = args.custom_metadata

    save_args = _fill_missing_save_or_restore_args(item, save_args, mode='save')
    byte_limiter = serialization.get_byte_limiter(self._save_concurrent_bytes)
    param_infos = self._get_param_infos(
        item,
        directory,
        use_ocdbt=self._use_ocdbt,
        ocdbt_target_data_file_size=ocdbt_target_data_file_size,
        enable_pinned_host_transfer=args.enable_pinned_host_transfer,
        byte_limiter=byte_limiter,
    )
    assert all(
        leaf.parent_dir == directory for leaf in jax.tree.leaves(param_infos)
    )

    serialize_ops = []  # List of (coros -> List of futures)
    batch_requests = batched_serialization_requests(
        item,
        param_infos,
        save_args,
        self._type_handler_registry,
    )
    tree_memory_size = 0
    for request in batch_requests:
      serialize_ops += [
          request.handler.serialize(request.values, request.infos, request.args)
      ]
      write_size, _ = _get_batch_memory_size(request.handler, request.values)
      tree_memory_size += write_size
    # Await copy futures. Returns List[List[future.Future]].
    commit_futures = await asyncio.gather(*serialize_ops)
    # Flatten to List[future.Future].
    commit_futures, _ = jax.tree.flatten(commit_futures)

    if logging.vlog_is_on(1):
      logging.vlog(1, 'param_info: %s', param_infos)
      logging.vlog(1, 'save_args: %s', save_args)

    save_futures = []
    if multihost.is_primary_host(self._primary_host):
      operation_id = (
          synchronization.HandlerAwaitableSignalOperationIdGenerator.get_current_operation_id()
      )
      save_futures.append(
          future.CommitFuture(
              self._write_metadata_after_commits(
                  commit_futures,
                  checkpoint_dir=directory,
                  param_infos=param_infos,
                  save_args=save_args,
                  custom_metadata=custom_metadata,
                  use_zarr3=self._use_zarr3,
                  operation_id=operation_id,
              ),
              name='write_metadata_after_commits',
          )
      )
    else:
      save_futures += commit_futures

    _log_io_metrics(
        tree_memory_size,
        start_time,
        '/jax/checkpoint/write/blocking_bytes_per_sec',
    )
    return [
        future.ChainedFuture(
            save_futures,
            functools.partial(
                _log_io_metrics,
                tree_memory_size,
                start_time,
                '/jax/checkpoint/write/bytes_per_sec',
                '/jax/checkpoint/write/bytes',
            ),
        )
    ]

  def save(self, directory: epath.Path, *args, **kwargs):
    """Saves the provided item.

    Blocks until both copy and commit complete.

    See async_save.

    Args:
      directory: the directory to save to.
      *args: additional arguments for save.
      **kwargs: additional arguments for save.
    """

    async def async_save(*args, **kwargs):
      commit_futures = await self.async_save(*args, **kwargs)  # pytype: disable=bad-return-type
      # Futures are already running, so sequential waiting is equivalent to
      # concurrent waiting.
      if commit_futures:  # May be None.
        for f in commit_futures:
          f.result()  # Block on result.

    asyncio_utils.run_sync(async_save(directory, *args, **kwargs))

  async def _maybe_deserialize(
      self,
      item: PyTree,
      metadata: PyTree,
      param_infos: PyTree,
      restore_args: PyTree,
  ) -> Tuple[int, PyTree]:
    """Deserializes values or skips."""
    flat_metadata = tree_utils.to_flat_dict(metadata)
    byte_limiter = serialization.get_byte_limiter(
        self._restore_concurrent_bytes
    )
    param_infos = jax.tree.map(
        lambda info: dataclasses.replace(info, byte_limiter=byte_limiter),
        param_infos,
    )
    batch_requests = batched_serialization_requests(
        metadata,
        param_infos,
        restore_args,
        self._type_handler_registry,
    )
    deserialized_batches = []
    deserialized_batches_ops = []
    for request in batch_requests:
      deserialized_batches_ops.append(
          request.handler.deserialize(request.infos, request.args)
      )
    deserialized_batches += await asyncio.gather(*deserialized_batches_ops)

    tree_memory_size = 0
    flat_restored = {}
    for request, deserialized in zip(batch_requests, deserialized_batches):
      _, read_size = _get_batch_memory_size(request.handler, deserialized)
      tree_memory_size += read_size
      for key, value in zip(request.keys, deserialized):
        flat_restored[key] = value
    # Add in empty nodes from the metadata tree.
    for key in flat_metadata.keys():
      if key not in flat_restored:
        flat_restored[key] = empty_values.get_empty_value_from_typestr(
            flat_metadata[key].value_type, self._pytree_metadata_options
        )
    # Restore using `item` as the target structure. If there are any custom
    # nodes (e.g. optax.EmptyState), these will replace None values in
    # flat_restored.
    return tree_memory_size, tree_utils.from_flat_dict(
        flat_restored, target=item
    )

  def restore(
      self,
      directory: epath.Path,
      args: Optional[BasePyTreeRestoreArgs] = None,
  ) -> PyTree:
    """Restores a PyTree from the checkpoint directory at the given path.

    In the most basic case, only `directory` is required. The tree will be
    restored exactly as saved, and all leaves will be restored as the correct
    types (assuming the tree metadata is present).

    However, `restore_args` is often required as well. This PyTree gives a
    `RestoreArgs` object (or subclass) for every leaf in the tree. Many types,
    such as string or `np.ndarray` do not require any special options for
    restoration. When restoring an individual leaf as `jax.Array`, however,
    some properties may be required.

    One example is `sharding`, which defines how a `jax.Array` in the restored
    tree should be partitioned. `mesh` and `mesh_axes` can also be used to
    specify `sharding`, but `sharding` is the preferred way of specifying this
    partition since `mesh` and `mesh_axes` only constructs
    `jax.sharding.NamedSharding`. For more information, see `ArrayTypeHandler`
    documentation and JAX sharding documentation.

    Example::

      ckptr = Checkpointer(BasePyTreeCheckpointHandler())
      restore_args = {
          'layer0': {
              'w': RestoreArgs(),
              'b': RestoreArgs(),
          },
          'layer1': {
              'w': ArrayRestoreArgs(
                  # Restores as jax.Array, regardless of how it was saved.
                  restore_type=jax.Array,
                  sharding=jax.sharding.Sharding(...),
                  # Warning: may truncate or pad!
                  global_shape=(x, y),
                ),
              'b': ArrayRestoreArgs(
                  restore_type=jax.Array,
                  sharding=jax.sharding.Sharding(...),
                  global_shape=(x, y),
                ),
          },
      }
      ckptr.restore(path, restore_args=restore_args)

    Providing `item` is typically only necessary when restoring a custom PyTree
    class (or when using transformations). In this case, the restored object
    will take on the same structure as `item`.

    Example::

      @flax.struct.dataclass
      class TrainState:
        layer0: dict[str, jax.Array]
        layer1: dict[str, jax.Array]

      ckptr = Checkpointer(BasePyTreeCheckpointHandler())
      train_state = TrainState(
          layer0={
              'w': jax.Array(...),  # zeros
              'b': jax.Array(...),  # zeros
          },
          layer1={
              'w': jax.Array(...),  # zeros
              'b': jax.Array(...),  # zeros
          },
      )
      restore_args = jax.tree.map(_make_restore_args, train_state)
      ckptr.restore(path, item=train_state, restore_args=restore_args)
      # restored tree is of type `TrainState`.

    Args:
      directory: saved checkpoint location directory.
      args: `BasePyTreeRestoreArgs` (see below).

    Returns:
      A PyTree matching the structure of `item`.

    Raises:
      FileNotFoundError: `directory` does not exist or is missing required files
      ValueError: `transforms` is provided without `item`.
      ValueError: `transforms` contains elements with `multi_value_fn`.
    """
    start_time = time.time()
    args = args or BasePyTreeRestoreArgs()
    item = args.item
    restore_args = args.restore_args

    logging.vlog(1, 'directory=%s, restore_args=%s', directory, restore_args)
    if not directory.exists():
      raise FileNotFoundError(
          f'Requested directory for restore does not exist at {directory}'
      )
    # Get value metadata tree and use_zarr3 from serialized pytree metadata.
    internal_tree_metadata = self._read_metadata_file(directory)
    value_metadata_tree = internal_tree_metadata.as_nested_tree()
    if not value_metadata_tree:
      raise ValueError(
          f'Found empty checkpoint PyTree metadata in directory={directory}.'
      )
    use_zarr3 = (
        internal_tree_metadata.use_zarr3
        if internal_tree_metadata.use_zarr3 is not None
        else self._use_zarr3
    )
    raise_array_data_missing_error = (
        internal_tree_metadata.store_array_data_equal_to_fill_value
    )
    del internal_tree_metadata
    # Prep for restore.
    if item is None:
      item = value_metadata_tree
    restore_args = _fill_missing_save_or_restore_args(
        item, restore_args, mode='restore'
    )
    restore_args = tree_metadata.serialize_tree(
        restore_args, self._pytree_metadata_options
    )
    param_infos = self._get_param_infos(
        item=value_metadata_tree,
        directory=directory,
        use_ocdbt=type_handlers.is_ocdbt_checkpoint(directory),
        use_zarr3=use_zarr3,
        raise_array_data_missing_error=raise_array_data_missing_error,
    )
    # Begin restore.
    tree_memory_size, restored_item = asyncio_utils.run_sync(
        self._maybe_deserialize(
            item, value_metadata_tree, param_infos, restore_args
        )
    )

    if logging.vlog_is_on(1):
      logging.vlog(1, 'param_infos: %s', param_infos)
      logging.vlog(1, 'checkpoint_restore_args: %s', restore_args)
      logging.vlog(1, 'restored_item: %s', jax.tree.structure(restored_item))
      logging.vlog(
          1,
          'ts_metrics: %s',
          json.dumps(ts.experimental_collect_matching_metrics('/tensorstore/')),
      )

    _log_io_metrics(
        tree_memory_size,
        start_time,
        '/jax/checkpoint/read/bytes_per_sec',
        '/jax/checkpoint/read/bytes',
    )
    return restored_item

  async def _get_param_infos_with_write_shape(
      self,
      param_infos: PyTree,
      checkpoint_dir: epath.Path,
      array_metadata_store: array_metadata_store_lib.Store,
  ) -> PyTree:
    """Returns `param_infos` updated with `write_shape`.

    Args:
      param_infos: A PyTree of ParamInfo to be updated.
      checkpoint_dir: The checkpoint directory where write_shape metadata is
        saved in ArrayMetadata store.
      array_metadata_store: The ArrayMetadata store to read write_shape metadata
        from.
    """
    if not utils.is_primary_host(self._primary_host):
      return param_infos
    # Extract write_shape from ArrayMetadata for current process_index.
    process_index = multihost.process_index()
    array_metadatas = await array_metadata_store.read(
        checkpoint_dir, process_index=process_index
    )
    if array_metadatas is None:
      jax_array_param_info = type_handlers.any_jax_array_param_info(param_infos)
      if jax_array_param_info is not None:
        raise ValueError(
            f'No ArrayMetadata found for process_index={process_index} in the'
            f' checkpoint directory: {checkpoint_dir}. But input PyTree'
            ' contains at least one jax.Array param_info:'
            f' {jax_array_param_info}.'
        )
      return param_infos

    assert isinstance(array_metadatas, list)
    array_metadatas_cache = {
        array_metadata.param_name: array_metadata
        for array_metadata in array_metadatas
    }

    def update_param_info(param_info: types.ParamInfo) -> types.ParamInfo:
      if not type_handlers.represents_jax_array(param_info):
        return param_info
      if param_info.name not in array_metadatas_cache:
        raise ValueError(
            f'No ArrayMetadata found for param_info: {param_info}, checkpoint'
            f' directory: {checkpoint_dir}, process_index={process_index}.'
        )
      return dataclasses.replace(
          param_info,
          write_shape=array_metadatas_cache[param_info.name].write_shape,
      )

    return jax.tree.map(update_param_info, param_infos)

  def _write_metadata_file(
      self,
      directory: epath.Path,
      *,
      param_infos: PyTree,
      save_args: PyTree,
      custom_metadata: tree_types.JsonType | None = None,
      use_zarr3: bool = False,
  ) -> future.Future:
    async def _save_fn(param_infos):
      if utils.is_primary_host(self._primary_host):
        metadata_write_start_time = time.time()
        path = directory / PYTREE_METADATA_FILE
        metadata_content = tree_metadata.InternalTreeMetadata.build(
            param_infos,
            save_args=save_args,
            use_zarr3=use_zarr3,
            custom_metadata=custom_metadata,
            pytree_metadata_options=self._pytree_metadata_options,
        )
        logging.vlog(
            1,
            'Writing pytree metadata file: %s with pytree_metadata_options: %s',
            path,
            self._pytree_metadata_options,
        )
        path.write_text(json.dumps(metadata_content.to_json()))
        jax.monitoring.record_event_duration_secs(
            '/jax/checkpoint/write/async/metadata_write_duration_secs',
            time.time() - metadata_write_start_time,
        )

    return future.CommitFuture(
        _save_fn(param_infos),
    )

  async def _write_metadata_after_commits(
      self,
      commit_futures: List[future.Future],
      checkpoint_dir: epath.Path,
      *,
      param_infos: PyTree,
      save_args: PyTree,
      custom_metadata: tree_types.JsonType | None = None,
      use_zarr3: bool,
      operation_id: str | None = None,
  ) -> None:
    if not utils.is_primary_host(self._primary_host):
      return
    for commit_future in commit_futures:
      await asyncio.to_thread(commit_future.result)
    # `write_shape` is extracted from ArrayMetadata store saved during
    # materialization of commit_futures. Then it is written to the pytree
    # metadata.
    # TODO(b/390465017): Simplify all metadata related code in this module after
    # removing overriding of self._write_metadata_file() in subclasses. All
    # metadata related code can be moved to a separate class and
    # BasePyTreeCheckpointHandler should delegate all metadata related code to
    # that class.
    if self._array_metadata_store is not None:
      param_infos = await self._get_param_infos_with_write_shape(
          param_infos, checkpoint_dir, self._array_metadata_store
      )

    async def metadata_coro():
      self._write_metadata_file(
          checkpoint_dir,
          param_infos=param_infos,
          save_args=save_args,
          custom_metadata=custom_metadata,
          use_zarr3=use_zarr3,
      ).result()

    metadata_write_future = future.CommitFutureAwaitingContractedSignals(
        metadata_coro(),
        name='write_metadata_file',
        operation_id=operation_id,
    )
    await asyncio.to_thread(metadata_write_future.result)

  def _read_metadata_file(
      self, directory: epath.Path
  ) -> tree_metadata.InternalTreeMetadata:
    """Reads metadata file and returns a tree of restore types.

    Args:
      directory: directory

    Returns:
      orbax.checkpoint.metadata.InternalTreeMetadata

    Raises:
      FileNotFoundError: if the metadata file is not found.
    """
    path = directory / PYTREE_METADATA_FILE
    if not path.exists():
      raise FileNotFoundError(
          f'Metadata file (named {PYTREE_METADATA_FILE}) does not exist at'
          f' {directory}.'
      )
    logging.vlog(
        1,
        'Reading pytree metadata file: %s with pytree_metadata_options: %s',
        path,
        self._pytree_metadata_options,
    )
    return tree_metadata.InternalTreeMetadata.from_json(
        json.loads(path.read_text()),
        pytree_metadata_options=self._pytree_metadata_options,
    )


  def metadata(self, directory: epath.Path) -> tree_metadata.TreeMetadata:
    """Returns tree metadata.

    The result will be a PyTree matching the structure of the saved checkpoint.
    Note that if the item saved was a custom class, the restored metadata will
    be returned as a nested dictionary representation.

    Example::

      {
        'layer0': {
            'w': ArrayMetadata(dtype=jnp.float32, shape=(8, 8), shards=(1, 2)),
            'b': ArrayMetadata(dtype=jnp.float32, shape=(8,), shards=(1,)),
        },
        'step': ScalarMetadata(dtype=jnp.int64),
      }

    If the required metadata file is not present, this method will raise an
    error.

    Args:
      directory: checkpoint location.

    Returns:
      tree containing metadata.
    """
    is_ocdbt_checkpoint = type_handlers.is_ocdbt_checkpoint(directory)
    internal_tree_metadata = self._read_metadata_file(directory)
    return tree_metadata.build_default_tree_metadata(
        internal_tree_metadata.as_custom_metadata(
            directory,
            self._type_handler_registry,
            use_ocdbt=is_ocdbt_checkpoint,
        ),
        custom_metadata=internal_tree_metadata.custom_metadata,
    )

  def finalize(self, directory: epath.Path) -> None:
    """Finalization step.

    Called automatically by the Checkpointer/AsyncCheckpointer just before the
    checkpoint is considered "finalized" in the sense of ensuring atomicity. See
    documentation for `type_handlers.merge_ocdbt_per_process_files`.

    Args:
      directory: Path where the checkpoint is located.
    """
    finalize_coros = []
    if self._array_metadata_store is not None:
      if self._primary_host is None:
        logging.log_first_n(
            logging.WARNING,
            '[process=%s] Skipped cross-host ArrayMetadata validation'
            ' because all hosts are primary (e.g. local storage).',
            1,  # log only once
            multihost.process_index(),
        )
      elif utils.is_primary_host(self._primary_host):
        finalize_coros.append(
            array_metadata_store_lib.validate_all_array_metadatas(
                self._array_metadata_validator,
                self._array_metadata_store,
                directory,
            )
        )

    async def merge_ocdbt_per_process_files():
      merge_start_time = time.time()
      ts_context = ts_utils.get_ts_context(use_ocdbt=True)
      await type_handlers.merge_ocdbt_per_process_files(
          directory,
          ts_context=ts_context,
          use_zarr3=self._use_zarr3,
          enable_validation=self._enable_post_merge_validation,
      )
      jax.monitoring.record_event_duration_secs(
          '/jax/checkpoint/write/async/ocdbt_merge_duration_secs',
          time.time() - merge_start_time,
      )

    finalize_coros.append(merge_ocdbt_per_process_files())

    async def _fn():
      await asyncio.gather(*finalize_coros)

    asyncio_utils.run_sync(_fn())


@register_with_handler(BasePyTreeCheckpointHandler, for_save=True)
@dataclasses.dataclass
class BasePyTreeSaveArgs(CheckpointArgs):
  """Parameters for saving a PyTree.

  Attributes:
    item (required): a PyTree to be saved.
    save_args: a PyTree with the same structure of `item`, which consists of
      `ocp.SaveArgs` objects as values. `None` can be used for values where no
      `SaveArgs` are specified.
    ocdbt_target_data_file_size: Specifies the target size (in bytes) of each
      OCDBT data file.  It only applies when OCDBT is enabled and Zarr3 must be
      turned on.  If left unspecified, default size is 2GB.  A value of 0
      indicates no maximum file size limit.  For best results, ensure
      chunk_byte_size is smaller than this value.  For more details, refer to
      https://google.github.io/tensorstore/kvstore/ocdbt/index.html#json-kvstore/ocdbt.target_data_file_size
    enable_pinned_host_transfer: If False, disables transfer to pinned host when
      copying from device to host, regardless of the presence of pinned host
      memory.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """

  item: PyTree
  save_args: Optional[PyTree] = None
  ocdbt_target_data_file_size: Optional[int] = None
  enable_pinned_host_transfer: bool = False
  custom_metadata: tree_types.JsonType | None = None


@register_with_handler(BasePyTreeCheckpointHandler, for_restore=True)
@dataclasses.dataclass
class BasePyTreeRestoreArgs(CheckpointArgs):
  """Parameters for restoring a PyTree.

  Attributes (all optional):
    item: provides the tree structure for the restored item. If not provided,
      will infer the structure from the saved checkpoint. Transformations will
      not be run in this case. Necessary particularly in the case where the
      caller needs to restore the tree as a custom object.
    restore_args: optional object containing additional arguments for
      restoration. It should be a PyTree matching the structure of `item`, or
      if `item` is not provided, then it should match the structure of the
      checkpoint. Each value in the tree should be a `RestoreArgs` object (OR
      a subclass of `RestoreArgs`). Importantly, note that when restoring a
      leaf as a certain type, a specific subclass of `RestoreArgs` may be
      required. `RestoreArgs` also provides the option to customize the
      restore type of an individual leaf.
  """

  item: Optional[PyTree] = None
  restore_args: Optional[PyTree] = None



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common test cases for PyTreeHandler."""

# pylint: disable=protected-access, missing-function-docstring

import asyncio
import contextlib
import copy
import dataclasses
import datetime
import functools
import json
import threading
from typing import Any, Awaitable, Iterator, List
from unittest import mock

from absl.testing import parameterized
from etils import epath
import flax
import flax.training.train_state
import jax
from jax import numpy as jnp
from jax.experimental import mesh_utils
import numpy as np
import optax
from orbax.checkpoint import test_utils
from orbax.checkpoint import utils
from orbax.checkpoint._src.arrays import abstract_arrays
from orbax.checkpoint._src.handlers import proto_checkpoint_handler
from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
from orbax.checkpoint._src.metadata import array_metadata
from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.metadata import sharding as sharding_metadata
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.metadata import value as value_metadata
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.serialization import replica_slices
from orbax.checkpoint._src.serialization import serialization
from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint._src.tree import utils as tree_utils
from orbax.checkpoint.experimental.v1._src.context import options
from orbax.checkpoint.experimental.v1._src.handlers import pytree_handler
from orbax.checkpoint.experimental.v1._src.testing import array_utils as array_test_utils
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


PyTree = tree_types.PyTree
ParamInfo = pytree_checkpoint_handler.ParamInfo
ArrayStorageOptions = options.ArrayStorageOptions

_SHARDING = '_sharding'
PYTREE_METADATA_FILE = pytree_checkpoint_handler.PYTREE_METADATA_FILE
ARRAY_METADATA_STORE = array_metadata_store_lib.Store()

create_sharded_array = array_test_utils.create_sharded_array
create_numpy_pytree = array_test_utils.create_numpy_pytree
create_sharded_pytree = array_test_utils.create_sharded_pytree
as_abstract_type = array_test_utils.as_abstract_type


async def _run_awaitable(awaitable: Awaitable[Any]) -> Any:
  return await awaitable


class PyTreeHandler:
  """Wrapper around PyTreeHandler that can block on save and load."""

  def __init__(self, **kwargs):
    self._handler = pytree_handler.PyTreeHandler(**kwargs)

  def save(self, *args, **kwargs):
    awaitable = asyncio.run(self._handler.save(*args, **kwargs))
    return asyncio.run(_run_awaitable(awaitable))

  def save_async(self, *args, **kwargs):
    return asyncio.run(self._handler.save(*args, **kwargs))

  def load(self, *args, **kwargs):
    awaitable = asyncio.run(self._handler.load(*args, **kwargs))
    return asyncio.run(_run_awaitable(awaitable))

  def load_async(self, *args, **kwargs):
    return asyncio.run(self._handler.load(*args, **kwargs))

  def metadata(self, *args, **kwargs):
    return asyncio.run(self._handler.metadata(*args, **kwargs))


def create_mixed_format_pytree(
    *,
    add: int = 0,
    strings: bool = False,
    parent_key: str | None = None,
    include_scalars: bool = True,
) -> PyTree:
  """Creates a PyTree with different leaf types for testing.

  Args:
    add: Adds the specified value to numeric leafs.
    strings: If true, adds string leaves to the tree.
    parent_key: If provided, keys will be contained within a dictionary under
      this key.
    include_scalars: If true, adds scalar leaves to the tree.

  Returns:
    PyTree
  """
  numpy_pytree, abstract_numpy_pytree = create_numpy_pytree(
      add=add, include_scalars=include_scalars
  )
  sharded_pytree, abstract_sharded_pytree = create_sharded_pytree(
      add=add, include_scalars=include_scalars
  )
  if parent_key:
    numpy_pytree = {parent_key: numpy_pytree}
    sharded_pytree = {parent_key: sharded_pytree}
    abstract_numpy_pytree = {parent_key: abstract_numpy_pytree}
    abstract_sharded_pytree = {parent_key: abstract_sharded_pytree}
  mixed_pytree = {
      'numpy': numpy_pytree,
      'sharded': sharded_pytree,
  }
  abstract_mixed_pytree = {
      'numpy': abstract_numpy_pytree,
      'sharded': abstract_sharded_pytree,
  }
  if strings:
    mixed_pytree['foo'] = 'foo_val'
    mixed_pytree['bar'] = 'bar_val'
    abstract_mixed_pytree['foo'] = ''
    abstract_mixed_pytree['bar'] = ''
  return mixed_pytree, abstract_mixed_pytree


def _raise_file_not_found_error(*args, **kwargs):
  del args, kwargs
  raise FileNotFoundError()


# Not in common util because we need to eliminate OSS dependency on flax.
def init_flax_model(model):
  params = model.init(jax.random.PRNGKey(0), jnp.ones([8, 8]))
  tx = optax.adamw(learning_rate=0.001)
  state = flax.training.train_state.TrainState.create(
      apply_fn=model.apply, params=params, tx=tx
  )
  return jax.tree.map(np.asarray, state)


@contextlib.contextmanager
def handler_with_options(
    *,
    create_array_storage_options_fn: (
        options.CreateArrayStorageOptionsFn | None
    ) = None,
    save_concurrent_bytes: int | None = None,
    restore_concurrent_bytes: int | None = None,
    use_ocdbt: bool = True,
    use_zarr3: bool = False,
    enable_padding_and_truncation: bool = True,
    ocdbt_target_data_file_size: int | None = None,
    enable_pinned_host_transfer: bool = False,
    pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
        tree_metadata.PYTREE_METADATA_OPTIONS
    ),
    array_metadata_store: array_metadata_store_lib.Store | None = (
        ARRAY_METADATA_STORE
    ),
):
  """Registers handlers with OCDBT support and resets when done."""
  type_handler_registry = copy.deepcopy(
      type_handlers.GLOBAL_TYPE_HANDLER_REGISTRY
  )
  type_handler_registry.get(jax.Array)._array_metadata_store = (
      array_metadata_store
  )

  handler = PyTreeHandler(
      create_array_storage_options_fn=create_array_storage_options_fn,
      save_concurrent_bytes=save_concurrent_bytes,
      restore_concurrent_bytes=restore_concurrent_bytes,
      use_ocdbt=use_ocdbt,
      use_zarr3=use_zarr3,
      enable_padding_and_truncation=enable_padding_and_truncation,
      ocdbt_target_data_file_size=ocdbt_target_data_file_size,
      enable_pinned_host_transfer=enable_pinned_host_transfer,
      type_handler_registry=type_handler_registry,
      pytree_metadata_options=pytree_metadata_options,
  )
  try:
    yield handler
  finally:
    pass


class PyTreeHandlerTestBase:
  """Base test cases for PyTreeCheckpointHandler."""

  class Test(parameterized.TestCase):
    """Test class."""

    def setUp(self):
      super().setUp()

      self.pytree, self.abstract_pytree = create_sharded_pytree()
      self.numpy_pytree, self.abstract_numpy_pytree = create_numpy_pytree()

      self.directory = epath.Path(
          self.create_tempdir(name='checkpointing_test').full_path
      )
      # TODO: b/365169723 - Add tests for support_rich_types=True.
      self.pytree_metadata_options = tree_metadata.PyTreeMetadataOptions(
          support_rich_types=False
      )


      # default to use_ocdbt=False, so we can test non-ocdbt handler first
      self.handler = self.enter_context(
          handler_with_options(
              use_ocdbt=False, array_metadata_store=ARRAY_METADATA_STORE
          )
      )
      test_utils.set_tensorstore_driver_for_test()

      test_utils.sync_global_processes(
          'PyTreeCheckpointHandlerTest:setup_complete'
      )

    def tearDown(self):
      test_utils.sync_global_processes(
          'PyTreeCheckpointHandlerTest:tests_complete'
      )
      super().tearDown()

    def validate_save(
        self,
        path: epath.Path,
        abstract_pytree: PyTree | None,
        expected: PyTree,
        checkpoint_handler: PyTreeHandler,
    ):
      """Validate save was performed correctly."""
      actual = checkpoint_handler.load(path, abstract_pytree)
      test_utils.assert_tree_equal(self, expected, actual)

    def validate_metadata(
        self,
        *,
        expected_reference_metadata_tree: PyTree,
        actual_metadata: PyTree,
        pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      """Validate metadata, provided the original tree that was saved."""
      expected_reference_metadata_tree = tree_metadata.serialize_tree(
          expected_reference_metadata_tree, pytree_metadata_options
      )

      def _metadata(value):
        if empty_values.is_supported_empty_value(
            value, pytree_metadata_options
        ):
          return value
        if isinstance(value, np.ndarray):
          return value_metadata.ArrayMetadata(
              name='',
              directory=None,
              shape=value.shape,
              sharding=None,
              dtype=value.dtype,
              storage=value_metadata.StorageMetadata(
                  chunk_shape=value.shape,
              ),
          )
        if isinstance(value, jax.Array):
          expected_sharding = sharding_metadata.from_jax_sharding(
              value.sharding
          )
          expected_chunk_shape = test_utils.get_expected_chunk_shape(value)
          return value_metadata.ArrayMetadata(
              name='',
              directory=None,
              shape=value.shape,
              sharding=expected_sharding,
              dtype=value.dtype,
              storage=value_metadata.StorageMetadata(
                  chunk_shape=expected_chunk_shape,
                  write_shape=(
                      expected_chunk_shape
                      if array_metadata_store is not None
                      else None
                  ),
              ),
          )
        if isinstance(value, (float, int)):
          dtype = np.float64 if isinstance(value, float) else np.int64
          return value_metadata.ScalarMetadata(
              name='', directory=None, dtype=dtype
          )  # pytype: disable=wrong-arg-types  # jnp-type
        if isinstance(value, str):
          return value_metadata.StringMetadata(name='', directory=None)
        if isinstance(value, optax.EmptyState):
          return None
        raise ValueError(f'Unrecognized type: {type(value)}.')

      expected_metadata = jax.tree.map(
          _metadata,
          expected_reference_metadata_tree,
          is_leaf=tree_utils.is_empty_or_leaf,
      )
      test_utils.assert_tree_equal(
          self, expected_metadata, actual_metadata.tree
      )

    def test_get_param_names(self):
      param_names = pytree_checkpoint_handler.get_param_names(self.pytree)
      expected = {
          'a': 'a',
          'b': 'b',
          'c': {
              'a': 'c.a',
              'e': 'c.e',
          },
          'x': 'x',
          'y': 'y',
      }
      test_utils.assert_tree_equal(self, expected, param_names)

    def test_save_format(self):
      pytree = {'a': 0, 'c': {'d': np.arange(3), 'e': {'f': 5}}, 'g': 10}
      self.handler.save(self.directory, pytree)
      fnames = ['a', 'c.d', 'c.e.f', 'g']
      paths = [self.directory / name for name in fnames]
      for p in paths:
        self.assertTrue(p.exists())
        self.assertTrue((p / '.zarray').exists())

    @parameterized.product(use_ocdbt=(True, False))
    def test_save_sharding(self, use_ocdbt: bool):
      if multihost.is_pathways_backend():
        self.skipTest('Sharding metadata not present on Pathways.')
      with handler_with_options(use_ocdbt=use_ocdbt) as checkpoint_handler:
        pytree = {
            'mlp/~/linear_0': {
                'a': self.pytree['a'],
                'b': self.pytree['b'],
                'c': {'a': self.pytree['c']['a'], 'e': self.pytree['c']['e']},
            }
        }
        abstract_pytree = jax.tree.map(
            array_test_utils.as_abstract_type, pytree
        )
        checkpoint_handler.save(self.directory, pytree)

        self.validate_save(
            self.directory,
            abstract_pytree,
            pytree,
            checkpoint_handler,
        )

      self.assertTrue((self.directory / _SHARDING).exists())
      with open(self.directory / _SHARDING, 'r') as file:
        data = json.load(file)
        self.assertCountEqual(
            data.keys(),
            {
                'bWxwL34vbGluZWFyXzAuYQ==',  # mlp/~/linear_0.a
                'bWxwL34vbGluZWFyXzAuYg==',  # mlp/~/linear_0.b
                'bWxwL34vbGluZWFyXzAuYy5h',  # mlp/~/linear_0.c.a
                'bWxwL34vbGluZWFyXzAuYy5l',  # mlp/~/linear_0.c.e
            },
        )
        # mlp/~/linear_0.a
        self.assertEqual(
            sharding_metadata.NamedShardingMetadata.from_deserialized_dict(
                json.loads(data['bWxwL34vbGluZWFyXzAuYQ=='])
            ),
            sharding_metadata.NamedShardingMetadata.from_jax_sharding(
                pytree['mlp/~/linear_0']['a'].sharding
            ),
        )
        # mlp/~/linear_0.b
        self.assertEqual(
            sharding_metadata.NamedShardingMetadata.from_deserialized_dict(
                json.loads(data['bWxwL34vbGluZWFyXzAuYg=='])
            ),
            sharding_metadata.NamedShardingMetadata.from_jax_sharding(
                pytree['mlp/~/linear_0']['b'].sharding
            ),
        )
        # mlp/~/linear_0.c.a
        self.assertEqual(
            sharding_metadata.NamedShardingMetadata.from_deserialized_dict(
                json.loads(data['bWxwL34vbGluZWFyXzAuYy5h'])
            ),
            sharding_metadata.NamedShardingMetadata.from_jax_sharding(
                pytree['mlp/~/linear_0']['c']['a'].sharding
            ),
        )
        # mlp/~/linear_0.c.e
        self.assertEqual(
            sharding_metadata.NamedShardingMetadata.from_deserialized_dict(
                json.loads(data['bWxwL34vbGluZWFyXzAuYy5l'])
            ),
            sharding_metadata.NamedShardingMetadata.from_jax_sharding(
                pytree['mlp/~/linear_0']['c']['e'].sharding
            ),
        )

    @parameterized.product(
        use_ocdbt=(True, False),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_disable_write_sharding_file(
        self,
        use_ocdbt: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      array_handler = type_handlers.ArrayHandler(
          enable_write_sharding_file=False,
          array_metadata_store=array_metadata_store,
      )
      ty = jax.Array
      fn = lambda ty: issubclass(ty, jax.Array)
      with test_utils.register_type_handler(ty, array_handler, fn):
        pytree, abstract_pytree = create_mixed_format_pytree()
        with handler_with_options(
            use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
        ) as checkpoint_handler:
          checkpoint_handler.save(self.directory, pytree)
          self.validate_save(
              self.directory,
              abstract_pytree,
              pytree,
              checkpoint_handler,
          )
      self.assertFalse((self.directory / _SHARDING).exists())

    def test_sharding_variable_devices(self):
      if multihost.is_pathways_backend():
        self.skipTest('Sharding metadata not present on Pathways.')
      mesh_axes = jax.sharding.PartitionSpec(
          'x',
      )
      devices_subset = []
      for idx in range(jax.process_count()):
        for d in jax.devices():
          if d.process_index == idx:
            devices_subset.append(d)
            break
      pytree = {
          'a': test_utils.create_sharded_array(
              np.arange(16),
              jax.sharding.Mesh(devices_subset, ('x',)),
              mesh_axes,
          ),
          'b': test_utils.create_sharded_array(
              np.arange(16), jax.sharding.Mesh(jax.devices(), ('x',)), mesh_axes
          ),
      }

      self.handler.save(self.directory, pytree)
      self.assertTrue((self.directory / _SHARDING).exists())
      a_sharding_metadata = sharding_metadata.NamedShardingMetadata(
          shape=np.array([2]),
          axis_names=['x'],
          partition_spec=('x',),
          device_mesh=sharding_metadata.DeviceMetadataMesh.from_jax_mesh(
              jax.sharding.Mesh(devices_subset, ('x',))
          ),
      )
      b_sharding_metadata = sharding_metadata.NamedShardingMetadata(
          shape=np.array([8]),
          axis_names=['x'],
          partition_spec=('x',),
          device_mesh=sharding_metadata.DeviceMetadataMesh.from_jax_mesh(
              jax.sharding.Mesh(jax.devices(), ('x',))
          ),
      )
      self.assertEqual(
          a_sharding_metadata,
          self.handler.metadata(self.directory)['a'].sharding,
      )
      self.assertEqual(
          b_sharding_metadata,
          self.handler.metadata(self.directory)['b'].sharding,
      )

    @parameterized.product(use_ocdbt=(True, False))
    def test_save_main(self, use_ocdbt: bool):
      with handler_with_options(use_ocdbt=use_ocdbt) as checkpoint_handler:
        checkpoint_handler.save(self.directory, self.pytree)
        self.validate_save(
            self.directory,
            self.abstract_pytree,
            self.pytree,
            checkpoint_handler,
        )
        self.assertEqual(
            type_handlers.is_ocdbt_checkpoint(self.directory), use_ocdbt
        )

    @parameterized.product(use_ocdbt=(True, False))
    def test_save_keys_with_slashes(self, use_ocdbt: bool):
      with handler_with_options(use_ocdbt=use_ocdbt) as checkpoint_handler:
        pytree = {
            'a': np.arange(2),
            'b/c': np.arange(4),
        }
        checkpoint_handler.save(self.directory, pytree)
        self.validate_save(
            self.directory,
            None,
            pytree,
            checkpoint_handler,
        )

    def test_save_non_sharded(self):
      self.handler.save(self.directory, self.numpy_pytree)
      self.validate_save(
          self.directory,
          None,
          self.numpy_pytree,
          self.handler,
      )

    @parameterized.product(
        use_ocdbt=(True, False),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_save_mixed(
        self,
        use_ocdbt: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      with handler_with_options(
          use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
      ) as checkpoint_handler:
        pytree, abstract_pytree = create_mixed_format_pytree(strings=True)
        checkpoint_handler.save(self.directory, pytree)
        self.validate_save(
            self.directory,
            abstract_pytree,
            pytree,
            checkpoint_handler,
        )
        if use_ocdbt:
          expected_files_and_directories = [
              '_strings.json',
              'manifest.ocdbt',
              'ocdbt.process_0',
          ]
        else:
          expected_files_and_directories = [
              '_strings.json',
              'numpy.a',
              'numpy.b',
              'numpy.c.a',
              'numpy.c.e',
          ]
        self.assertContainsSubset(
            expected_files_and_directories,
            [f.name for f in self.directory.iterdir()],
        )
        self.validate_metadata(
            expected_reference_metadata_tree=pytree,
            actual_metadata=checkpoint_handler.metadata(self.directory),
            pytree_metadata_options=self.pytree_metadata_options,
            array_metadata_store=array_metadata_store,
        )

    @parameterized.product(
        use_ocdbt=(True, False),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_save_strings(
        self,
        use_ocdbt: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      if use_ocdbt and multihost.is_pathways_backend():
        self.skipTest('Pathways + OCDBT not supported.')

      with handler_with_options(
          use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
      ) as checkpoint_handler:
        pytree, abstract_pytree = create_mixed_format_pytree(strings=True)

        checkpoint_handler.save(self.directory, pytree)
        self.validate_save(
            self.directory,
            abstract_pytree,
            pytree,
            checkpoint_handler,
        )
        self.validate_metadata(
            expected_reference_metadata_tree=pytree,
            actual_metadata=checkpoint_handler.metadata(self.directory),
            pytree_metadata_options=self.pytree_metadata_options,
            array_metadata_store=array_metadata_store,
        )
        self.assertTrue((self.directory / '_strings.json').exists())
        with open(self.directory / '_strings.json') as file:
          data = json.load(file)
          self.assertCountEqual(
              data.keys(),
              {'foo', 'bar'},
              None,
          )
          self.assertEqual(data['foo'], 'foo_val')
          self.assertEqual(data['bar'], 'bar_val')

    def test_cast(self):
      pytree, abstract_pytree = create_mixed_format_pytree(
          include_scalars=False
      )
      origin_dtype = np.int64
      save_dtype = np.uint32
      restore_dtype = np.float64

      def check_dtype(x, dtype):
        if not utils.is_scalar(x):
          self.assertEqual(x.dtype, dtype)

      def set_dtype(v, dtype):
        if hasattr(v, 'dtype'):
          setattr(v, 'dtype', dtype)
        return v

      with self.subTest('check_origin_dtype'):
        jax.tree.map(functools.partial(check_dtype, dtype=origin_dtype), pytree)
        jax.tree.map(
            functools.partial(check_dtype, dtype=origin_dtype), abstract_pytree
        )

      with handler_with_options(
          use_ocdbt=False,
          create_array_storage_options_fn=lambda k, v: ArrayStorageOptions(
              dtype=save_dtype
          ),
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, pytree)

      with self.subTest('check_restore_dtype'):
        abstract_pytree = jax.tree.map(
            functools.partial(set_dtype, dtype=restore_dtype), abstract_pytree
        )
        restored = self.handler.load(self.directory, abstract_pytree)
        jax.tree.map(
            functools.partial(check_dtype, dtype=restore_dtype), restored
        )

      with self.subTest('check_save_dtype'):
        restored = self.handler.load(self.directory)
        jax.tree.map(functools.partial(check_dtype, dtype=save_dtype), restored)

    def test_cast_scalar(self):
      pytree = {'a': 5, 'b': 1.2}
      abstract_pytree = {'a': 0.0, 'b': 0}
      self.handler.save(self.directory, pytree)
      restored = self.handler.load(self.directory, abstract_pytree)
      self.assertIsInstance(restored['a'], float)
      self.assertIsInstance(restored['b'], int)

    def test_load_type(self):
      pytree = {'a': 5, 'b': 6.1}
      abstract_pytree = {
          'a': np.asarray(0.0, dtype=np.float32),
          'b': np.asarray(0, dtype=np.int32),
      }

      self.handler.save(self.directory, pytree)
      restored = self.handler.load(self.directory, abstract_pytree)
      self.assertIsInstance(restored['a'], np.ndarray)
      self.assertIsInstance(restored['b'], np.ndarray)
      self.assertEqual(restored['a'].dtype, np.float32)
      self.assertEqual(restored['b'].dtype, np.int32)

    @parameterized.product(
        use_ocdbt=(True, False),
        use_zarr3=(True, False),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_save_restore(
        self,
        use_ocdbt: bool,
        use_zarr3: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      with handler_with_options(
          use_ocdbt=use_ocdbt,
          use_zarr3=use_zarr3,
          array_metadata_store=array_metadata_store,
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, self.pytree)
        restored = checkpoint_handler.load(
            self.directory,
            self.abstract_pytree,
        )
        test_utils.assert_tree_equal(self, self.pytree, restored)
        self.validate_metadata(
            expected_reference_metadata_tree=self.pytree,
            actual_metadata=checkpoint_handler.metadata(self.directory),
            pytree_metadata_options=self.pytree_metadata_options,
            array_metadata_store=array_metadata_store,
        )

    def test_save_async(self):
      save_done = threading.Event()
      original_finalize = pytree_handler.PyTreeHandler._finalize

      def mock_finalize(*args, **kwargs):
        save_done.wait()  # Wait for explicit signal before proceeding.
        return original_finalize(*args, **kwargs)

      def is_save_complete(directory):
        return (directory / 'manifest.ocdbt').exists()

      with mock.patch.object(
          pytree_handler.PyTreeHandler,
          '_finalize',
          new=mock_finalize,
      ), handler_with_options() as checkpoint_handler:
        awaitable = checkpoint_handler.save_async(self.directory, self.pytree)
        self.assertFalse(is_save_complete(self.directory))
        save_done.set()
        asyncio.run(_run_awaitable(awaitable))
        self.assertTrue(is_save_complete(self.directory))
        restored = checkpoint_handler.load(
            self.directory,
            self.abstract_pytree,
        )
        test_utils.assert_tree_equal(self, self.pytree, restored)

    def test_load_async(self):
      with handler_with_options() as checkpoint_handler:
        checkpoint_handler.save(self.directory, self.pytree)
        load_awaitable = checkpoint_handler.load_async(
            self.directory,
            self.abstract_pytree,
        )
        restored = asyncio.run(_run_awaitable(load_awaitable))
        test_utils.assert_tree_equal(self, self.pytree, restored)

    @parameterized.product(use_ocdbt=(True, False))
    def test_load_reverse_mesh(self, use_ocdbt: bool):
      if use_ocdbt and multihost.is_pathways_backend():
        self.skipTest('Pathways + OCDBT not supported.')
      with handler_with_options(use_ocdbt=use_ocdbt) as checkpoint_handler:
        pytree, abstract_pytree = array_test_utils.create_sharded_pytree(
            reverse_devices=True
        )
        checkpoint_handler.save(self.directory, pytree)
        restored = checkpoint_handler.load(self.directory, abstract_pytree)
        test_utils.assert_tree_equal(self, pytree, restored)

    def test_load_multiple_steps(self):
      for step in [0, 1]:
        directory = self.directory / str(step)
        if multihost.process_index() == 0:
          directory.mkdir()
        test_utils.sync_global_processes(
            'PyTreeCheckpointHandlerTest:test_load_different_mkdir'
        )

        pytree, abstract_pytree = create_mixed_format_pytree(add=step)
        self.handler.save(directory, pytree)

        restored = self.handler.load(directory, abstract_pytree)
        test_utils.assert_tree_equal(self, pytree, restored)

    def test_load_missing_checkpoint(self):
      directory = self.directory / 'nothing'
      with self.assertRaises(FileNotFoundError):
        self.handler.load(directory)

    @parameterized.product(
        use_ocdbt=(True, False),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_flax_model(
        self,
        use_ocdbt: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):

      @flax.struct.dataclass
      class Params(flax.struct.PyTreeNode):
        params: PyTree
        opt_state: PyTree

      def make_state_with_optax():
        return Params(
            params=self.numpy_pytree,
            opt_state=(optax.EmptyState(), optax.EmptyState()),
        )

      def make_state_with_nones():
        return Params(
            params=self.numpy_pytree,
            opt_state=(None, None),
        )

      state = make_state_with_optax()

      with handler_with_options(
          use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, state)

        with self.subTest('with_abstract_state'):
          abstract_state = jax.tree.map(
              array_test_utils.as_abstract_type, state
          )
          restored = checkpoint_handler.load(self.directory, abstract_state)
          expected_state = state
          test_utils.assert_tree_equal(self, expected_state, restored)
          self.validate_metadata(
              expected_reference_metadata_tree=expected_state,
              actual_metadata=checkpoint_handler.metadata(self.directory),
              pytree_metadata_options=self.pytree_metadata_options,
              array_metadata_store=array_metadata_store,
          )

        with self.subTest('without_abstract_state'):
          if multihost.is_pathways_backend():
            self.skipTest('Must provide abstract_pytree for Pathways.')
          restored = checkpoint_handler.load(self.directory)
          expected_state = tree_utils.serialize_tree(
              make_state_with_nones(),
              keep_empty_nodes=True,
          )
          test_utils.assert_tree_equal(self, expected_state, restored)
          self.validate_metadata(
              expected_reference_metadata_tree=expected_state,
              actual_metadata=checkpoint_handler.metadata(self.directory),
              pytree_metadata_options=self.pytree_metadata_options,
              array_metadata_store=array_metadata_store,
          )

    @parameterized.product(
        use_ocdbt=(
            True,
            False,
        ),
        data=(
            {},
            {'a': {}, 'b': 3},
            [1, {}, 2],
            None,
            {'a': None, 'b': 3},
            [1, None, 2],
            [],
            [1, [], 2],
            {'a': [], 'b': 3},
        ),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_empty_data(
        self,
        use_ocdbt: bool,
        data: Any,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      with handler_with_options(
          use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
      ) as checkpoint_handler:
        if not data:
          with self.assertRaisesRegex(ValueError, 'Found empty item'):
            checkpoint_handler.save(
                self.directory,
                data,
            )
          return

        checkpoint_handler.save(self.directory, data)
        restored = checkpoint_handler.load(self.directory)
        self.assertEqual(restored, data)

        self.validate_metadata(
            expected_reference_metadata_tree=data,
            actual_metadata=checkpoint_handler.metadata(self.directory),
            pytree_metadata_options=self.pytree_metadata_options,
            array_metadata_store=array_metadata_store,
        )

    @parameterized.product(
        use_ocdbt=(True, False),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_list(
        self,
        use_ocdbt: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      item = [1, 2, 5, 6]
      with handler_with_options(
          use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, item)
        abstract_item = [0, 0, 0, 0]
        restored = checkpoint_handler.load(self.directory, abstract_item)
        self.assertListEqual(restored, item)
        self.validate_metadata(
            expected_reference_metadata_tree=[0, 0, 0, 0],
            actual_metadata=checkpoint_handler.metadata(self.directory),
            pytree_metadata_options=self.pytree_metadata_options,
            array_metadata_store=array_metadata_store,
        )

        restored = checkpoint_handler.load(self.directory)
        self.assertListEqual(
            restored,
            [
                np.asarray([1]),
                np.asarray([2]),
                np.asarray([5]),
                np.asarray([6]),
            ],
        )

    def test_no_metadata_file(self):
      self.handler.save(self.directory, self.pytree)
      metadata_file = self.directory / PYTREE_METADATA_FILE
      if multihost.process_index() == 0:
        self.assertTrue(metadata_file.exists())
        metadata_file.unlink()
      test_utils.sync_global_processes('delete_metadata_file')
      self.assertFalse(metadata_file.exists())
      with self.assertRaises(FileNotFoundError):
        self.handler.metadata(self.directory)

    @parameterized.product(
        use_ocdbt=(False,),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_override_tensorstore_metadata_name_sharded_arrays(
        self,
        use_ocdbt: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      metadata_key = 'custom_zarray'
      ty = jax.Array
      array_handler = type_handlers.ArrayHandler(
          metadata_key=metadata_key, array_metadata_store=array_metadata_store
      )
      fn = lambda ty: issubclass(ty, jax.Array)
      pytree = self.pytree

      with test_utils.register_type_handler(ty, array_handler, fn):
        with handler_with_options(
            use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
        ) as checkpoint_handler:
          checkpoint_handler.save(self.directory, pytree)
          param_names = pytree_checkpoint_handler.get_param_names(pytree)
          paths = jax.tree.map(lambda n: self.directory / n, param_names)

          def check_path(p):
            self.assertTrue(
                (p / metadata_key).exists(),
                msg=f'{p / metadata_key} should exist.',
            )
            self.assertFalse(
                (p / '.zarray').exists(),
                msg=f'{p / ".zarray"} should not exist.',
            )

          jax.tree.map(check_path, paths)

    @parameterized.product(
        use_ocdbt=(False,),
        array_metadata_store=(None, ARRAY_METADATA_STORE),
    )
    def test_override_tensorstore_metadata_name(
        self,
        use_ocdbt: bool,
        array_metadata_store: array_metadata_store_lib.Store | None,
    ):
      metadata_key = 'custom_zarray'
      ty = np.ndarray
      np_array_handler = type_handlers.NumpyHandler(metadata_key=metadata_key)
      fn = lambda ty: issubclass(ty, np.ndarray)
      pytree = self.numpy_pytree
      del pytree['x']
      del pytree['y']

      with test_utils.register_type_handler(ty, np_array_handler, fn):
        with handler_with_options(
            use_ocdbt=use_ocdbt, array_metadata_store=array_metadata_store
        ) as checkpoint_handler:
          checkpoint_handler.save(self.directory, pytree)
          param_names = pytree_checkpoint_handler.get_param_names(pytree)
          paths = jax.tree.map(lambda n: self.directory / n, param_names)

          def check_path(p):
            self.assertTrue(
                (p / metadata_key).exists(),
                msg=f'{p / metadata_key} should exist.',
            )
            self.assertFalse(
                (p / '.zarray').exists(),
                msg=f'{p / ".zarray"} should not exist.',
            )

          jax.tree.map(check_path, paths)

    @parameterized.parameters((True,), (False,))
    def test_reshape_padding(self, enable_padding_and_truncation: bool):
      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('x',))
      axes = jax.sharding.PartitionSpec(
          'x',
      )
      dtype = np.float32
      pytree = {
          'x': test_utils.create_sharded_array(
              np.arange(8, dtype=dtype), mesh, axes
          )
      }
      abstract_pytree = {
          'x': jax.ShapeDtypeStruct(
              shape=(16,), dtype=dtype, sharding=pytree['x'].sharding
          )
      }
      with handler_with_options(
          enable_padding_and_truncation=enable_padding_and_truncation
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, pytree)
        if enable_padding_and_truncation:
          restored = checkpoint_handler.load(self.directory, abstract_pytree)
          expected = {
              'x': test_utils.create_sharded_array(
                  np.concatenate(
                      (np.arange(8, dtype=dtype), np.zeros(8, dtype=dtype))
                  ),
                  mesh,
                  axes,
              )
          }
          test_utils.assert_tree_equal(self, expected, restored)
        else:
          with self.assertRaises(BaseException):
            checkpoint_handler.load(self.directory, abstract_pytree)

    @parameterized.parameters((True,), (False,))
    def test_reshape_truncate(self, enable_padding_and_truncation: bool):
      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('x',))
      axes = jax.sharding.PartitionSpec(
          'x',
      )
      dtype = np.float32
      pytree = {
          'x': test_utils.create_sharded_array(
              np.arange(16, dtype=dtype), mesh, axes
          )
      }
      abstract_pytree = {
          'x': jax.ShapeDtypeStruct(
              shape=(8,), dtype=dtype, sharding=pytree['x'].sharding
          )
      }

      with handler_with_options(
          enable_padding_and_truncation=enable_padding_and_truncation
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, pytree)
        if enable_padding_and_truncation:
          restored = checkpoint_handler.load(self.directory, abstract_pytree)
          expected = {
              'x': test_utils.create_sharded_array(
                  np.arange(8, dtype=dtype), mesh, axes
              )
          }
          test_utils.assert_tree_equal(self, expected, restored)
        else:
          with self.assertRaises(BaseException):
            checkpoint_handler.load(self.directory, abstract_pytree)

    @parameterized.parameters(
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('x', 'y'))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('y', 'x'))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('x',))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('y',))),
        (jax.sharding.PartitionSpec(('x', 'y')), jax.sharding.PartitionSpec()),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('x',)),
        ),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('y',)),
        ),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('y', 'x')),
        ),
        (
            jax.sharding.PartitionSpec(('x',)),
            jax.sharding.PartitionSpec(('y',)),
        ),
    )
    def test_reshard(self, save_spec, restore_spec):
      devices = jax.devices()
      len_devices = len(devices)
      self.assertGreaterEqual(len_devices, 4)

      mesh = jax.sharding.Mesh(
          mesh_utils.create_device_mesh((4, len_devices // 4)), ('x', 'y')
      )
      dtype = np.int32
      pytree = {
          'x': test_utils.create_sharded_array(
              np.arange(len_devices, dtype=dtype), mesh, save_spec
          )
      }
      abstract_pytree = {
          'x': jax.ShapeDtypeStruct(
              shape=(len_devices,),
              dtype=dtype,
              sharding=jax.sharding.NamedSharding(mesh, restore_spec),
          )
      }

      self.handler.save(self.directory, pytree)
      restored = self.handler.load(self.directory, abstract_pytree)
      expected = {
          'x': test_utils.create_sharded_array(
              np.arange(len_devices, dtype=dtype), mesh, restore_spec
          )
      }
      test_utils.assert_tree_equal(self, expected, restored)

    def test_load_non_ocdbt(self):
      with handler_with_options(use_ocdbt=False) as checkpoint_handler:
        checkpoint_handler.save(self.directory, self.pytree)
        self.assertFalse(type_handlers.is_ocdbt_checkpoint(self.directory))
      with handler_with_options(use_ocdbt=True) as checkpoint_handler:
        restored = checkpoint_handler.load(
            self.directory,
            self.abstract_pytree,
        )
        test_utils.assert_tree_equal(self, self.pytree, restored)

    def test_load_non_ocdbt_mixed(self):
      pytree, abstract_pytree = create_mixed_format_pytree(strings=True)
      with handler_with_options(use_ocdbt=False) as checkpoint_handler:
        checkpoint_handler.save(self.directory, pytree)
        self.assertFalse(type_handlers.is_ocdbt_checkpoint(self.directory))
      with handler_with_options(use_ocdbt=True) as checkpoint_handler:
        restored = checkpoint_handler.load(self.directory, abstract_pytree)
        test_utils.assert_tree_equal(self, pytree, restored)

    def test_check_zarray(self):
      self.handler.save(self.directory, self.pytree)
      zarr_path = self.directory / 'a' / '.zarray'
      zarr_path.unlink(missing_ok=True)
      test_utils.sync_global_processes(
          'PyTreeCheckpointHandlerTest:delete_zarray'
      )
      self.assertFalse(zarr_path.exists())
      with self.assertRaises(FileNotFoundError):
        self.handler.load(
            self.directory,
            self.abstract_pytree,
        )

    def test_without_abstract_pytree(self):
      arr = test_utils.create_sharded_array(
          np.arange(8),
          jax.sharding.Mesh(jax.devices(), ('x',)),
          jax.sharding.PartitionSpec('x'),
      )
      pytree = [arr]
      self.handler.save(self.directory, pytree)
      restored = self.handler.load(self.directory)
      test_utils.assert_tree_equal(self, pytree, restored)

    @parameterized.product(use_ocdbt=(True, False))
    def test_masked_shape_dtype_struct(self, use_ocdbt: bool):

      def _should_mask(keypath):
        return keypath[0].key == 'a' or (
            keypath[0].key == 'c' and keypath[1].key == 'e'
        )

      def _mask(keypath, x):
        return optax.MaskedNode() if _should_mask(keypath) else x

      def _none(keypath, x):
        return None if _should_mask(keypath) else x

      masked_tree = jax.tree_util.tree_map_with_path(_mask, self.pytree)
      expected = jax.tree_util.tree_map_with_path(_none, self.pytree)

      with handler_with_options(use_ocdbt=use_ocdbt) as handler:
        handler.save(self.directory, masked_tree)
        if use_ocdbt:
          self.assertTrue(type_handlers.is_ocdbt_checkpoint(self.directory))

        # Restore it with state which was given before applying masking.
        restored = handler.load(
            self.directory,
            jax.tree.map(abstract_arrays.to_shape_dtype_struct, self.pytree),
        )
        test_utils.assert_tree_equal(self, expected, restored)

        # Restore it with state after applying masking to it.
        restored = handler.load(
            self.directory,
            jax.tree.map(abstract_arrays.to_shape_dtype_struct, masked_tree),
        )
        test_utils.assert_tree_equal(self, expected, restored)

        # Restore it without any state.
        restored = handler.load(
            self.directory,
            self.abstract_pytree,
        )
        test_utils.assert_tree_equal(self, expected, restored)

    def test_finalize(self):
      with handler_with_options(use_ocdbt=True) as checkpoint_handler:
        checkpoint_handler.save(self.directory, self.pytree)
        process_index = multihost.process_index()
        process_dir = (
            self.directory / f'{ts_utils.PROCESS_SUBDIR_PREFIX}{process_index}'
        )
        self.assertTrue(process_dir.exists())
        self.assertTrue(process_dir.is_dir())
        self.assertTrue(type_handlers.is_ocdbt_checkpoint(self.directory))

    @parameterized.product(use_ocdbt=(True, False))
    def test_unregistered_types(self, use_ocdbt: bool):
      data = {'uncheckpointable_field': datetime.timedelta(seconds=5)}
      with handler_with_options(use_ocdbt=use_ocdbt) as checkpoint_handler:
        with self.assertRaisesRegex(ValueError, 'TypeHandler lookup failed'):
          checkpoint_handler.save(
              self.directory,
              data,
          )

    @parameterized.product(
        target_data_file_size=[
            50 * 1024,  # 50KB
            10 * 1024,  # 10KB
            0,
            None,
        ],
        chunk_byte_size=[
            None,  # unspecified
            5 * 1024,  # 5KB
            100 * 1024,  # greater than target_data_file_size
        ],
        use_zarr3=[True, False],
        patch_default_ocdbt_data_file_size=[True, False],
    )
    def test_ocdbt_target_data_file_size(
        self,
        target_data_file_size,
        chunk_byte_size,
        use_zarr3,
        patch_default_ocdbt_data_file_size,
    ):
      """Test ocdbt_target_data_file_size."""
      array_len = 16 * 1024  # ~ 64KB of float data
      custom_pytree = {
          'a': np.arange(array_len, dtype=np.int32),
          'b': np.arange(array_len * 2, dtype=np.float32),
          'c': {
              'a': np.arange(array_len, dtype=np.int32).reshape(
                  2, array_len // 2
              ),
              'e': np.arange(array_len * 2, dtype=np.float32).reshape(
                  2, array_len
              ),
          },
      }
      shardings = {
          'a': self.abstract_pytree['a'].sharding,
          'b': self.abstract_pytree['b'].sharding,
          'c': {
              'a': self.abstract_pytree['c']['a'].sharding,
              'e': self.abstract_pytree['c']['e'].sharding,
          },
      }
      pytree = jax.tree.map(create_sharded_array, custom_pytree, shardings)
      abstract_pytree = jax.tree.map(as_abstract_type, pytree)

      create_array_storage_options_fn = (
          lambda key, value: options.ArrayStorageOptions(
              chunk_byte_size=chunk_byte_size
          )
      )

      new_ocdbt_target_data_file_size = (
          1024
          if patch_default_ocdbt_data_file_size
          else ts_utils._DEFAULT_OCDBT_TARGET_DATA_FILE_SIZE
      )
      with mock.patch.object(
          ts_utils,
          '_DEFAULT_OCDBT_TARGET_DATA_FILE_SIZE',
          new_ocdbt_target_data_file_size,
      ):
        if patch_default_ocdbt_data_file_size:
          assert ts_utils._DEFAULT_OCDBT_TARGET_DATA_FILE_SIZE == 1024
        with handler_with_options(
            use_ocdbt=True,
            use_zarr3=use_zarr3,
            ocdbt_target_data_file_size=target_data_file_size,
            create_array_storage_options_fn=create_array_storage_options_fn,
        ) as checkpoint_handler:
          checkpoint_handler.save(self.directory, pytree)

          data_dir = self.directory / 'd'
          self.assertTrue(data_dir.exists())
          self.assertTrue(data_dir.is_dir())

          for f in data_dir.iterdir():
            if f.is_file():
              if target_data_file_size not in (0, None):
                # it's expected the resulting file sizes can be larger than
                # the target_data_file_size, so we give some buffer here
                self.assertLessEqual(
                    f.stat().length,
                    target_data_file_size * 2.0,
                )  # some room
                if patch_default_ocdbt_data_file_size:
                  self.assertLessEqual(
                      f.stat().length,
                      new_ocdbt_target_data_file_size * 2.0,
                  )

          restored = checkpoint_handler.load(self.directory, abstract_pytree)

          test_utils.assert_tree_equal(self, pytree, restored)

    def test_local_registry(self):

      if multihost.is_pathways_backend():
        # This does not test anything on the pathways backend
        # TODO(b/333114195): add proper pathways testing.
        return

      class PlusOneHandler(type_handlers.ScalarHandler):

        async def serialize(self, values, infos, args=None):
          values = [v + 1 for v in values]
          return await super().serialize(values, infos, args)

      registry = type_handlers.create_type_handler_registry(
          (int, PlusOneHandler()),
      )
      handler = PyTreeHandler(type_handler_registry=registry)
      with self.assertRaisesRegex(
          ValueError, "TypeHandler lookup failed for: type=<class 'float'>"
      ):
        handler.save(self.directory, {'a': 3, 'b': 1.0})
      handler.save(self.directory, {'a': 3})

      restored = handler.load(self.directory)
      expected = {'a': 4}
      self.assertEqual(restored, expected)


    def test_empty_custom_node(self):

      class PyTreeDict(dict):
        pass

      jax.tree_util.register_pytree_node(
          PyTreeDict,
          lambda d: (tuple(d.values()), tuple(d.keys())),
          lambda keys, values: PyTreeDict(dict(zip(keys, values))),
      )

      with self.assertRaisesRegex(ValueError, 'Found empty item'):
        self.handler.save(self.directory, PyTreeDict())

      self.handler.save(self.directory, {'a': PyTreeDict()})
      restored = self.handler.load(self.directory)
      self.assertDictEqual({'a': {}}, restored)

      restored = self.handler.load(self.directory, {'a': PyTreeDict()})
      test_utils.assert_tree_equal(self, {'a': PyTreeDict()}, restored)

    @parameterized.parameters((5,), (9,))
    def test_concurrent_gb_save(self, limit_bytes):
      # TODO(b/346811105): Enable for Pathways.
      if multihost.is_pathways_backend():
        self.skipTest(
            'Disabled on Pathways because completion_times cannot updated by'
            ' reference outside remote Python.'
        )
      sleep_time = 1.0
      sharding = jax.sharding.NamedSharding(
          jax.sharding.Mesh(
              jax.devices(),
              ('x',),
          ),
          jax.sharding.PartitionSpec(
              None,
          ),
      )
      # 4 arrays, each has a single chunk, with 4 bytes each.
      tree = jax.tree.map(
          functools.partial(
              array_test_utils.create_sharded_array, sharding=sharding
          ),
          {
              'a': np.arange(1, dtype=np.int32),
              'b': np.arange(1, dtype=np.int32),
              'c': np.arange(1, dtype=np.int32),
              'd': np.arange(1, dtype=np.int32),
          },
      )
      byte_limiter = test_utils.get_byte_limiter(limit_bytes, sleep_time)
      with mock.patch.object(
          serialization,
          'get_byte_limiter',
          new=lambda _: byte_limiter,
      ), handler_with_options(
          save_concurrent_bytes=limit_bytes,
      ) as handler:
        handler.save(self.directory, tree)
      # Replicated shards are handled within the _write_array_shard function.
      # Since shards are only saved once per replica, we only have to check
      # the primary process.
      completion_times = byte_limiter.completion_times
      if multihost.process_index() == 0:
        self.assertLen(completion_times, len(jax.tree.leaves(tree)))
        test_utils.assert_every_n_is_x_apart(
            self,
            completion_times,
            limit_bytes // np.int32().itemsize,
            sleep_time,
        )

    @parameterized.parameters((5,), (9,))
    def test_concurrent_gb_restore(self, limit_bytes):
      # TODO(b/346811105): Enable for Pathways.
      if multihost.is_pathways_backend():
        self.skipTest(
            'Disabled on Pathways because completion_times cannot updated by'
            ' reference outside remote Python.'
        )
      sleep_time = 1.0
      sharding = jax.sharding.NamedSharding(
          jax.sharding.Mesh(
              jax.devices(),
              ('x',),
          ),
          jax.sharding.PartitionSpec(
              None,
          ),
      )
      # 4 arrays, each has a single chunk, with 4 bytes each.
      tree = jax.tree.map(
          functools.partial(
              array_test_utils.create_sharded_array, sharding=sharding
          ),
          {
              'a': np.arange(1, dtype=np.int32),
              'b': np.arange(1, dtype=np.int32),
              'c': np.arange(1, dtype=np.int32),
              'd': np.arange(1, dtype=np.int32),
          },
      )
      self.handler.save(self.directory, tree)

      byte_limiter = test_utils.get_byte_limiter(limit_bytes, sleep_time)
      with mock.patch.object(
          serialization,
          'get_byte_limiter',
          new=lambda _,: byte_limiter,
      ), handler_with_options(restore_concurrent_bytes=limit_bytes) as handler:
        restored = handler.load(self.directory)
      test_utils.assert_tree_equal(self, tree, restored)
      completion_times = byte_limiter.completion_times
      self.assertLen(
          completion_times,
          len(jax.tree.leaves(tree)),
      )
      test_utils.assert_every_n_is_x_apart(
          self,
          completion_times,
          limit_bytes // np.int32().itemsize,
          sleep_time,
      )

    def test_enable_pinned_host_transfer(self):
      if multihost.is_pathways_backend():
        self.skipTest(
            'Disabled on Pathways because local variables cannot updated by'
            ' reference outside remote Python.'
        )
      true_count = 0
      false_count = 0

      original_transfer_arrays_to_host = replica_slices.transfer_arrays_to_host

      def _transfer_arrays_to_host(
          arrays, replica_id, use_replica_parallel, enable_pinned_host_transfer
      ):
        nonlocal true_count, false_count
        if enable_pinned_host_transfer:
          true_count += 1
        else:
          false_count += 1
        return original_transfer_arrays_to_host(
            arrays,
            replica_id,
            use_replica_parallel=use_replica_parallel,
            enable_pinned_host_transfer=enable_pinned_host_transfer,
        )

      with mock.patch.object(
          replica_slices,
          'transfer_arrays_to_host',
          new=_transfer_arrays_to_host,
      ), handler_with_options(enable_pinned_host_transfer=False) as handler:
        handler.save(self.directory, self.pytree)

      self.assertEqual(true_count, 0)
      self.assertGreater(false_count, 0)

    @parameterized.product(
        use_ocdbt=(True, False),
        pytree_metadata_options=(
            tree_metadata.PyTreeMetadataOptions(support_rich_types=False),
            tree_metadata.PyTreeMetadataOptions(support_rich_types=True),
        ),
    )
    def test_write_shape_metadata_missing_for_all_types_other_than_jax_array(
        self,
        use_ocdbt: bool,
        pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
    ):
      checkpoint = {
          'a': 1,
          'b': np.array([2]),
          'c': 'hello',
      }
      expected_metadata_without_directory = {
          'a': value_metadata.ScalarMetadata(
              name='a',
              directory=None,
              shape=(),
              sharding=None,
              dtype=np.dtype('int64'),
              storage=None,
          ),
          'b': value_metadata.ArrayMetadata(
              name='b',
              directory=None,
              shape=(1,),
              sharding=None,
              dtype=np.dtype('int64'),
              storage=value_metadata.StorageMetadata(
                  chunk_shape=(1,), write_shape=None
              ),
          ),
          'c': value_metadata.StringMetadata(name='c', directory=None),
      }
      with handler_with_options(
          use_ocdbt=use_ocdbt,
          pytree_metadata_options=pytree_metadata_options,
          array_metadata_store=ARRAY_METADATA_STORE,
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, checkpoint)

        self.assertFalse((self.directory / 'array_metadatas').exists())
        metadata = checkpoint_handler.metadata(self.directory)
        metadata_without_directory = jax.tree.map(
            lambda m: dataclasses.replace(m, directory=None), metadata.tree
        )
        self.assertEqual(
            expected_metadata_without_directory,
            metadata_without_directory,
        )

    @parameterized.product(
        use_ocdbt=(True, False),
        pytree_metadata_options=(
            tree_metadata.PyTreeMetadataOptions(support_rich_types=False),
            tree_metadata.PyTreeMetadataOptions(support_rich_types=True),
        ),
    )
    def test_write_shape_in_metadata_disabled(
        self,
        use_ocdbt: bool,
        pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
    ):
      with handler_with_options(
          use_ocdbt=use_ocdbt,
          pytree_metadata_options=pytree_metadata_options,
          array_metadata_store=None,
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, self.pytree)
        expected_tree_with_write_shapes = {
            'a': {'write_shape': None},
            'b': {'write_shape': None},
            'c': {
                'a': {'write_shape': None},
                'e': {'write_shape': None},
            },
            'x': {'write_shape': None},
            'y': {'write_shape': None},
        }
        metadata = checkpoint_handler.metadata(self.directory)
        tree_with_write_shapes = jax.tree.map(
            lambda m: {'write_shape': m.storage.write_shape}, metadata.tree
        )
        self.assertDictEqual(
            expected_tree_with_write_shapes, tree_with_write_shapes
        )

    # TODO(b/382230550): Add test for chunk_shape != write_shape.
    @parameterized.product(
        use_ocdbt=(True, False),
        pytree_metadata_options=(
            tree_metadata.PyTreeMetadataOptions(support_rich_types=False),
            tree_metadata.PyTreeMetadataOptions(support_rich_types=True),
        ),
    )
    def test_write_shape_in_metadata(
        self,
        use_ocdbt: bool,
        pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
    ):
      with handler_with_options(
          use_ocdbt=use_ocdbt, pytree_metadata_options=pytree_metadata_options
      ) as checkpoint_handler:
        checkpoint_handler.save(self.directory, self.pytree)

        expected_tree_with_write_shapes = {
            'a': {'write_shape': (8,)},
            'b': {'write_shape': (2,)},
            'c': {
                'a': {'write_shape': (1, 1)},
                'e': {'write_shape': (2, 1)},
            },
            'x': {'write_shape': ()},
            'y': {'write_shape': ()},
        }
        metadata = checkpoint_handler.metadata(self.directory)
        tree_with_write_shapes = jax.tree.map(
            lambda m: {'write_shape': m.storage.write_shape}, metadata.tree
        )
        self.assertDictEqual(
            expected_tree_with_write_shapes, tree_with_write_shapes
        )

    @parameterized.product(use_ocdbt=(True, False))
    def test_array_metadata_disabled(self, use_ocdbt: bool):
      with handler_with_options(
          use_ocdbt=use_ocdbt, array_metadata_store=None
      ) as checkpoint_handler:
        pytree, abstract_pytree = create_mixed_format_pytree()

        checkpoint_handler.save(self.directory, pytree)

        self.validate_save(
            self.directory,
            abstract_pytree,
            pytree,
            checkpoint_handler,
        )

      self.assertFalse((self.directory / 'array_metadatas').exists())

    @parameterized.product(use_ocdbt=(True, False))
    def test_array_metadata(self, use_ocdbt: bool):
      with handler_with_options(use_ocdbt=use_ocdbt) as checkpoint_handler:

        checkpoint_handler.save(self.directory, self.pytree)

        self.validate_save(
            self.directory,
            self.abstract_pytree,
            self.pytree,
            checkpoint_handler,
        )

      self.assertTrue((self.directory / 'array_metadatas').exists())
      if multihost.process_index() == 0:
        array_metadatas = asyncio.run(ARRAY_METADATA_STORE.read(self.directory))
        self.assertIsInstance(array_metadatas, dict)
        per_process_metadatas = [
            array_metadata.SerializedArrayMetadata(
                param_name='a',
                write_shape=(8,),
                chunk_shape=(8,),
            ),
            array_metadata.SerializedArrayMetadata(
                param_name='b',
                write_shape=(2,),
                chunk_shape=(2,),
            ),
            array_metadata.SerializedArrayMetadata(
                param_name='c.a',
                write_shape=(1, 1),
                chunk_shape=(1, 1),
            ),
            array_metadata.SerializedArrayMetadata(
                param_name='c.e',
                write_shape=(2, 1),
                chunk_shape=(2, 1),
            ),
            array_metadata.SerializedArrayMetadata(
                param_name='x',
                write_shape=(),
                chunk_shape=(),
            ),
            array_metadata.SerializedArrayMetadata(
                param_name='y',
                write_shape=(),
                chunk_shape=(),
            ),
        ]
        expected_array_metadatas = {
            idx: per_process_metadatas
            for idx in range(multihost.process_count())
        }
        self.assertSameElements(
            expected_array_metadatas.keys(), array_metadatas.keys()
        )
        for process_index in expected_array_metadatas:
          self.assertEqual(  # pylint: disable=g-generic-assert
              sorted(
                  expected_array_metadatas[process_index],
                  key=lambda x: x.param_name,
              ),
              sorted(
                  array_metadatas[process_index], key=lambda x: x.param_name
              ),
          )

    @parameterized.product(use_ocdbt=(True, False))
    def test_save_with_missing_array_metadata_file(self, use_ocdbt: bool):
      if multihost.process_index() != 0:  # only test on primary host
        self.skipTest('Test only for primary host to avoid barrier timeout.')

      class PathResolverReturningNoMetadataFiles(
          array_metadata_store_lib.PathResolver
      ):

        def get_read_file_paths(
            self, checkpoint_dir: epath.Path, process_index: int | None = None
        ) -> Iterator[epath.Path] | epath.Path | None:
          return None

      with handler_with_options(
          use_ocdbt=use_ocdbt,
          array_metadata_store=array_metadata_store_lib.Store(
              path_resolver=PathResolverReturningNoMetadataFiles()
          ),
      ) as checkpoint_handler:
        with self.assertRaisesRegex(
            ValueError, 'No ArrayMetadata found for process_index'
        ):
          checkpoint_handler.save(self.directory, self.pytree)

    @parameterized.product(use_ocdbt=(True, False))
    def test_save_with_missing_array_metadata_for_params(self, use_ocdbt: bool):
      if multihost.process_index() != 0:  # only test on primary host
        self.skipTest('Test only for primary host to avoid barrier timeout.')

      class MissingArrayMetadataSerDeserializer(
          array_metadata_store_lib.SerDeserializer
      ):

        def deserialize(
            self, serialized: str
        ) -> List[array_metadata.SerializedArrayMetadata]:
          true_data = super().deserialize(serialized)
          return [true_data.pop(0)]  # Delete the rest and return partial data.

      with handler_with_options(
          use_ocdbt=use_ocdbt,
          array_metadata_store=array_metadata_store_lib.Store(
              ser_deser=MissingArrayMetadataSerDeserializer()
          ),
      ) as checkpoint_handler:
        with self.assertRaisesRegex(
            ValueError, 'No ArrayMetadata found for param_info'
        ):
          checkpoint_handler.save(self.directory, self.pytree)

    @parameterized.parameters((True,), (False,))
    def test_zero_size_array(self, use_jax_array: bool):
      arr = np.ones(shape=(0,))
      mesh = jax.sharding.Mesh(np.array(jax.devices()), axis_names=('x',))
      pspec = jax.sharding.PartitionSpec()
      if use_jax_array:
        arr = test_utils.create_sharded_array(arr, mesh, pspec)
      pytree = [arr]
      with self.assertRaisesRegex(ValueError, 'zero size'):
        self.handler.save(self.directory, pytree)

    @parameterized.product(use_ocdbt=(True, False))
    def test_save_restore_random_keys(self, use_ocdbt: bool):
      """Test saving and restoring random keys within a pytree."""

      # TODO(b/393160483) investigate Pathways remote Python support for
      # random.keys.
      if multihost.is_pathways_backend():
        self.skipTest(
            'Disabled on Pathways because random keys are not supported by'
            ' remote Python.'
        )

      mesh = jax.sharding.Mesh(jax.devices(), ('x',))
      sharding = jax.sharding.NamedSharding(mesh, jax.sharding.PartitionSpec())

      pytree = {
          'keys': {
              'kone': jax.random.key(jnp.array(0, device=sharding)),
              'impl_key': {
                  'rbg': jax.random.key(
                      jnp.array(1, device=sharding), impl='rbg'
                  ),
                  'unsafe_rbg': jax.random.key(
                      jnp.array(2, device=sharding), impl='unsafe_rbg'
                  ),
              },
              'split_keys': jax.random.split(
                  jax.random.key(jnp.array(123, device=sharding)), num=10
              ),
          },
          'arrays': self.pytree,
      }

      with handler_with_options(
          use_ocdbt=use_ocdbt,
          array_metadata_store=array_metadata_store_lib.Store(),
      ) as save_handler:
        save_handler.save(self.directory, pytree)

      with handler_with_options(
          use_ocdbt=use_ocdbt,
          array_metadata_store=array_metadata_store_lib.Store(),
      ) as load_handler:
        restored = load_handler.load(self.directory)
        test_utils.assert_tree_equal(self, pytree, restored)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of `CheckpointableHandler` for PyTrees."""

import dataclasses
from typing import Awaitable, Sequence

from etils import epath
import jax
import numpy as np
from orbax.checkpoint import checkpoint_utils
from orbax.checkpoint import options as options_lib
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.futures import synchronization
from orbax.checkpoint._src.handlers import base_pytree_checkpoint_handler
from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint.experimental.v1._src.context import options
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types

PathLike = path_types.PathLike
AsyncResponse = async_types.AsyncResponse
CheckpointableHandler = handler_types.CheckpointableHandler
PyTree = tree_types.PyTree


def _get_legacy_compatible_save_args(
    checkpointable: PyTree,
    create_array_storage_options_fn: options.CreateArrayStorageOptionsFn | None,
) -> PyTree | None:
  """Returns save args that are compatible with the V0 API."""
  if create_array_storage_options_fn is None:
    return None

  def _leaf_get_legacy_compatible_save_args(k, v):
    array_storage_options = create_array_storage_options_fn(k, v)
    save_dtype = (
        np.dtype(array_storage_options.dtype)
        if array_storage_options.dtype
        else None
    )
    return type_handlers.SaveArgs(
        dtype=save_dtype,
        chunk_byte_size=array_storage_options.chunk_byte_size,
        shard_axes=array_storage_options.shard_axes,
    )

  return jax.tree.map_with_path(
      _leaf_get_legacy_compatible_save_args, checkpointable
  )


class PyTreeHandler(
    CheckpointableHandler[
        PyTree,
        PyTree,
        PyTree,
    ]
):
  """An implementation of `CheckpointableHandler` for PyTrees."""

  # TODO(b/398249409): Currently `PyTreeHandler` is not used by higher-level
  # code. Many of the options here need to be moved to the `Context`.
  def __init__(
      self,
      *,
      create_array_storage_options_fn: (
          options.CreateArrayStorageOptionsFn | None
      ) = None,
      save_concurrent_bytes: int | None = None,
      restore_concurrent_bytes: int | None = None,
      use_ocdbt: bool = True,
      use_zarr3: bool = True,
      enable_padding_and_truncation: bool = False,
      ocdbt_target_data_file_size: int | None = None,
      enable_pinned_host_transfer: bool = False,
      multiprocessing_options: options_lib.MultiprocessingOptions = options_lib.MultiprocessingOptions(),
      type_handler_registry: type_handlers.TypeHandlerRegistry = type_handlers.GLOBAL_TYPE_HANDLER_REGISTRY,
      enable_post_merge_validation: bool = True,
      pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
          tree_metadata.PYTREE_METADATA_OPTIONS
      ),
      array_metadata_validator: array_metadata_store_lib.Validator = (
          array_metadata_store_lib.Validator()
      ),
  ):
    self._handler_impl = base_pytree_checkpoint_handler.BasePyTreeCheckpointHandler(
        save_concurrent_bytes=save_concurrent_bytes,
        restore_concurrent_bytes=restore_concurrent_bytes,
        use_ocdbt=use_ocdbt,
        use_zarr3=use_zarr3,
        multiprocessing_options=multiprocessing_options,
        type_handler_registry=type_handler_registry,
        enable_post_merge_validation=enable_post_merge_validation,
        pytree_metadata_options=pytree_metadata_options,
        array_metadata_validator=array_metadata_validator,
    )
    self._create_array_storage_options_fn = create_array_storage_options_fn
    self._ocdbt_target_data_file_size = ocdbt_target_data_file_size
    self._enable_pinned_host_transfer = enable_pinned_host_transfer
    self._multiprocessing_options = multiprocessing_options
    self._enable_padding_and_truncation = enable_padding_and_truncation

  def _finalize(self, directory: path_types.Path):
    if multihost.is_primary_host(self._multiprocessing_options.primary_host):
      self._handler_impl.finalize(directory)

  async def _background_save(
      self,
      directory: path_types.Path,
      *,
      commit_futures: Sequence[future.Future],
      operation_id: str,
  ):
    for f in commit_futures:
      f.result()
    # Global sync to ensure all participating processes have completed their
    # save operations before proceeding to finalize.
    barrier_name = f'save_and_finalize_{operation_id}_commit_complete'
    multihost.sync_global_processes(
        barrier_name, processes=self._multiprocessing_options.active_processes
    )
    # Finalize.
    self._finalize(directory)
    # Global sync to ensure all hosts are aware that the finalize operation
    # has completed before returning to the user.
    barrier_name = f'save_and_finalize_{operation_id}_finalize_complete'
    multihost.sync_global_processes(
        barrier_name, processes=self._multiprocessing_options.active_processes
    )

  async def save(
      self, directory: path_types.PathLike, checkpointable: PyTree
  ) -> Awaitable[None]:
    directory = epath.Path(directory)
    commit_futures = await self._handler_impl.async_save(
        directory,
        args=base_pytree_checkpoint_handler.BasePyTreeSaveArgs(
            item=checkpointable,
            save_args=_get_legacy_compatible_save_args(
                checkpointable, self._create_array_storage_options_fn
            ),
            ocdbt_target_data_file_size=self._ocdbt_target_data_file_size,
            enable_pinned_host_transfer=self._enable_pinned_host_transfer,
        ),
    )
    assert commit_futures

    # TODO(b/398310070): Move operation ID generation to `Context`.
    operation_id = (
        synchronization.HandlerAwaitableSignalOperationIdGenerator.get_current_operation_id()
    )
    return self._background_save(
        directory, commit_futures=commit_futures, operation_id=operation_id
    )

  async def _background_load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: PyTree | None = None,
  ) -> PyTree:

    def _set_enable_padding_and_truncation(a):
      if not isinstance(a, type_handlers.ArrayRestoreArgs):
        return a
      return dataclasses.replace(
          a,
          strict=not self._enable_padding_and_truncation,
      )

    restore_args = checkpoint_utils.construct_restore_args(
        abstract_checkpointable
    )
    restore_args = jax.tree.map(
        _set_enable_padding_and_truncation, restore_args
    )
    args = base_pytree_checkpoint_handler.BasePyTreeRestoreArgs(
        item=abstract_checkpointable,
        restore_args=restore_args,
    )
    return self._handler_impl.restore(directory, args=args)

  async def load(
      self,
      directory: path_types.PathLike,
      abstract_checkpointable: PyTree | None = None,
  ) -> Awaitable[PyTree]:
    directory = epath.Path(directory)
    return self._background_load(directory, abstract_checkpointable)

  async def metadata(self, directory: path_types.PathLike) -> PyTree:
    directory = epath.Path(directory)
    return self._handler_impl.metadata(directory)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for serializing and deserializing metadata."""

from typing import Any, Sequence


def validate_type(obj: Any, field_type: type[Any] | Sequence[type[Any]]):
  if isinstance(field_type, Sequence):
    if not any(isinstance(obj, f_type) for f_type in field_type):
      raise ValueError(
          f'Object must be any one of types {list(field_type)}, got '
          f'{type(obj)}.'
      )
  elif not isinstance(obj, field_type):
    raise ValueError(f'Object must be of type {field_type}, got {type(obj)}.')


def validate_and_process_custom_metadata(
    custom_metadata: Any,
) -> dict[str, Any]:
  """Validates and processes custom field."""
  if custom_metadata is None:
    return {}

  validate_type(custom_metadata, dict)
  for k in custom_metadata:
    validate_type(k, str)
  return custom_metadata




# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ReplicatorCheckpointManager for emergency checkpoint. See details below.

WARNING: Do not use without specific approval. The API and implementation are
subject to change without notice.
"""

import dataclasses
from typing import Any, Callable, Iterable, List, Sequence, Tuple

from absl import logging
from etils import epath
from etils import epy
import jax
from orbax.checkpoint import abstract_checkpoint_manager
from orbax.checkpoint import args as args_lib
from orbax.checkpoint import checkpoint_manager
from orbax.checkpoint._src import composite
from orbax.checkpoint._src.handlers import handler_registration
from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint.experimental.emergency import mesh_consistency
from orbax.checkpoint.experimental.emergency import process_metadata_checkpoint_handler
from orbax.checkpoint.path import step as step_lib
from typing_extensions import Self  # for Python version < 3.11


PyTree = Any
DefaultCheckpointHandlerRegistry = (
    handler_registration.DefaultCheckpointHandlerRegistry
)
PyTreeCheckpointHandler = pytree_checkpoint_handler.PyTreeCheckpointHandler
RootMetadata = checkpoint_manager.RootMetadata
StepMetadata = checkpoint_manager.StepMetadata
ProcessMetadataCheckpointHandler = (
    process_metadata_checkpoint_handler.ProcessMetadataCheckpointHandler
)


_UNNAMED_ITEM_NAME = 'state'
_PROCESS_METADATA_NAME = 'process_metadata'


def _local_checkpoint_handler(
    multiprocessing_options: checkpoint_manager.MultiprocessingOptions,
) -> Tuple[PyTreeCheckpointHandler, ProcessMetadataCheckpointHandler]:
  """Create a PyTreeCheckpointHandler for local checkpoints."""
  if multiprocessing_options.primary_host is not None:
    raise ValueError(
        'multiprocessing_options.primary_host must be set to None for local'
        ' checkpoints.'
    )
  local_registry = type_handlers.create_type_handler_registry(
      (
          jax.Array,
          type_handlers.ArrayHandler(
              primary_host=None, replica_id=None, use_replica_parallel=False
          ),
      ),
  )
  pytree_handler = PyTreeCheckpointHandler(
      use_ocdbt=True,
      use_zarr3=True,
      multiprocessing_options=multiprocessing_options,
      type_handler_registry=local_registry,
  )
  metadata_handler = ProcessMetadataCheckpointHandler()
  return pytree_handler, metadata_handler


@dataclasses.dataclass
class ReplicatorCheckpointManagerOptions:
  save_interval_steps: int = 1
  step_name_format: step_lib.NameFormat[step_lib.Metadata] | None = None
  should_save_fn: Callable[[int, int | None], bool] | None = None


class ReplicatorCheckpointManager(
    abstract_checkpoint_manager.AbstractCheckpointManager, epy.ContextManager
):
  """ReplicatorCheckpointManager.

  This class is intended for use in emergency checkpointing, but the system must
  conform to restrictive assumptions in order to work. Consider using
  `emergency.checkpoint_manager.CheckpointManager` if these assumptions are not
  met.

  This class assumes that an independent service is responsible for reflecting
  any checkpoints saved in local storage (e.g. RAMFS) to a persistent global
  storage (e.g. GCS). Thus, this class only saves checkpoints to process-local
  storage. Each process saves addressable data to its own storage location, on
  every process.

  When restoring a checkpoint, this class assumes that the independent
  replicator service will ensure consistency across all process-local storages.
  This means that if a restart causes data loss on one or more processes, the
  replicator must ensure that the corresponding data from a process with intact
  data is copied to the process with data loss. Specifically, the processes must
  be "peers" in the sense that they share the same index into the global array
  data.

  Users can control a few properties of the checkpointing behavior like the save
  interval. The period at which checkpoints are persisted to global storage, or
  the period at which they are
  garbage-collected, must be controlled by the replicator service.
  """

  def __init__(
      self,
      local_directory: epath.Path,
      options: ReplicatorCheckpointManagerOptions | None = None,
      *,
      global_mesh: jax.sharding.Mesh,
  ):
    self._global_mesh = global_mesh
    options = options or ReplicatorCheckpointManagerOptions()
    multiprocessing_options = checkpoint_manager.MultiprocessingOptions(
        primary_host=None
    )
    self._options = options
    self._step_name_format = (
        options.step_name_format or step_lib.standard_name_format()
    )
    self._options = dataclasses.replace(
        self._options,
        step_name_format=self._step_name_format,
    )
    [state_handler, process_metadata_handler] = _local_checkpoint_handler(
        multiprocessing_options
    )
    options = checkpoint_manager.CheckpointManagerOptions(
        save_interval_steps=options.save_interval_steps,
        max_to_keep=None,  # No garbage collection.
        step_name_format=options.step_name_format,
        should_save_fn=options.should_save_fn,
        multiprocessing_options=multiprocessing_options,
        create=False,
        cleanup_tmp_directories=False,  # Handled separately below.
        enable_background_delete=False,
        enable_async_checkpointing=True,
    )

    self._handler_registry = DefaultCheckpointHandlerRegistry()
    self._handler_registry.add(None, args_lib.PyTreeSave, state_handler)
    self._handler_registry.add(None, args_lib.PyTreeRestore, state_handler)
    self._handler_registry.add(
        None,
        process_metadata_checkpoint_handler.ProcessMetadataSaveArgs,
        process_metadata_handler,
    )
    self._handler_registry.add(
        None,
        process_metadata_checkpoint_handler.ProcessMetadataRestoreArgs,
        process_metadata_handler,
    )
    self._impl = checkpoint_manager.CheckpointManager(
        local_directory,
        options=options,
        handler_registry=self._handler_registry,
    )
    self._run_initial_garbage_collection()

  def _run_initial_garbage_collection(self):
    """Remove steps that might be left over from previous runs."""
    logging.info('Running initial garbage collection at %s.', self.directory)
    logging.info('Cleaning up existing temporary directories.')
    tmp_files = step_lib.tmp_checkpoints(self.directory)
    logging.info('Found tmp files: %s', tmp_files)
    for tmp_file in tmp_files:
      if (
          tmp_file
          == mesh_consistency.process_metadata_folder(self.directory).name
      ):
        continue
      (self.directory / tmp_file).rmtree()

  @property
  def directory(self) -> epath.Path:
    return self._impl.directory

  @property
  def global_mesh(self) -> jax.sharding.Mesh:
    return self._global_mesh

  def all_steps(self, read: bool = False) -> Sequence[int]:
    return self._impl.all_steps(read=read)

  def latest_step(self) -> int | None:
    return self._impl.latest_step()

  def best_step(self) -> int | None:
    raise NotImplementedError()

  def reload(self):
    return self._impl.reload()

  def reached_preemption(self, step: int) -> bool:
    return self._impl.reached_preemption(step)

  def should_save(self, step: int) -> bool:
    return self._impl.should_save(step)

  def delete(self, step: int):
    return self._impl.delete(step)

  def _validate_and_standardize_args(
      self,
      args: args_lib.CheckpointArgs | None,
  ) -> args_lib.Composite:
    if not isinstance(args, args_lib.CheckpointArgs):
      raise ValueError(
          f'Expected args to be a `CheckpointArgs`, but got {type(args)}.'
      )
    if not isinstance(args, composite.Composite):
      args = args_lib.Composite(**{_UNNAMED_ITEM_NAME: args})
    for a in args.values():
      assert isinstance(a, args_lib.CheckpointArgs)
      if not self._handler_registry.has(None, a):
        raise ValueError(
            f'{type(a)} is not supported by this CheckpointManager. This is'
            ' likely because it does not yet implement support for local'
            ' checkpointing.'
        )
    return args

  def save(
      self,
      step: int,
      args: args_lib.CheckpointArgs,
      *,
      force: bool = False,
  ) -> bool:
    args = self._validate_and_standardize_args(args)
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'CheckpointManager:save_start',
            prefix='replicator_checkpoint_manager',
        ),
        record_event_name=(
            '/jax/checkpoint/write/checkpoint_start_sync_duration_secs'
        ),
    )
    process_metadata_args = (
        process_metadata_checkpoint_handler.ProcessMetadataSaveArgs(
            global_mesh=self._global_mesh,
        )
    )
    args_dict = dict(args.items())
    args_dict[_PROCESS_METADATA_NAME] = process_metadata_args
    args = args_lib.Composite(**args_dict)
    saved = self._impl.save(step, args=args, force=force)
    return saved

  def _get_mesh_consistent_args(
      self,
      previous_distributed_to_device_ids: List[List[int]],
      previous_device_ids: List[int],
      args: args_lib.Composite,
  ) -> tuple[args_lib.Composite, args_lib.Composite]:
    restore_mesh = mesh_consistency.consistent_restore_mesh_from_metadata(
        self._global_mesh,
        multihost.distributed_to_device_ids(),
        previous_distributed_to_device_ids=previous_distributed_to_device_ids,
        previous_device_ids=previous_device_ids,
    )

    def _replace_sharding(arg: type_handlers.ArrayRestoreArgs):
      if arg.sharding is None:
        raise ValueError('ArrayRestoreArgs sharding cannot be None.')
      if not isinstance(arg.sharding, jax.sharding.NamedSharding):
        raise ValueError(
            'ArrayRestoreArgs sharding must be a NamedSharding, but got'
            f' {type(arg.sharding)}.'
        )
      return dataclasses.replace(
          arg,
          sharding=jax.sharding.NamedSharding(
              mesh=restore_mesh, spec=arg.sharding.spec
          ),
      )

    original_args = args
    consistent_args = {}
    for k, a in args.items():
      if isinstance(a, args_lib.PyTreeRestore):
        if a.restore_args is None:
          raise ValueError('`restore_args` cannot be None.')
        consistent_args[k] = dataclasses.replace(
            a, restore_args=jax.tree.map(_replace_sharding, a.restore_args)
        )
    return original_args, args_lib.Composite(**consistent_args)

  def _get_mesh_consistent_result(
      self,
      original_args: args_lib.Composite,
      consistent_result: args_lib.Composite,
      *,
      default_item_mode: bool,
  ):
    result = {}
    for k, a in original_args.items():
      item_result = consistent_result[k]
      if isinstance(a, args_lib.PyTreeRestore):
        original_shardings = jax.tree.map(
            lambda arg: arg.sharding, a.restore_args
        )
        item_result = mesh_consistency.consistent_restore_mesh_to_global_mesh(
            item_result, original_shardings
        )
      result[k] = item_result

    result = args_lib.Composite(**result)
    if default_item_mode:
      assert len(result) == 1
      return result[_UNNAMED_ITEM_NAME]
    else:
      return result

  def restore(
      self,
      step: int | None,
      args: args_lib.CheckpointArgs | None = None,
  ) -> Any:
    if step is None:
      step = self.latest_step()
      if step is None:
        raise FileNotFoundError(f'No steps found in {self.directory}.')
    default_item_mode = (
        checkpoint_manager.determine_default_item_mode_from_args(args)
    )
    args = self._validate_and_standardize_args(args)
    process_meatadata_args = args_lib.Composite(**{
        _PROCESS_METADATA_NAME: (
            process_metadata_checkpoint_handler.ProcessMetadataRestoreArgs()
        )
    })
    process_metadata_restored = self._impl.restore(
        step, args=process_meatadata_args
    )
    previous_distributed_to_device_ids, previous_device_ids = (
        process_metadata_restored[_PROCESS_METADATA_NAME]
    )
    original_args, consistent_args = self._get_mesh_consistent_args(
        previous_distributed_to_device_ids, previous_device_ids, args
    )
    restored = self._impl.restore(step, args=consistent_args)
    return self._get_mesh_consistent_result(
        original_args, restored, default_item_mode=default_item_mode
    )

  def item_metadata(self, step: int) -> Any:
    return self._impl.item_metadata(step)

  def metadata(self, step: int | None = None) -> RootMetadata | StepMetadata:
    return self._impl.metadata(step)

  def metrics(self, step: int) -> PyTree | None:
    raise NotImplementedError()

  def wait_until_finished(self):
    return self._impl.wait_until_finished()

  def check_for_errors(self):
    return self._impl.check_for_errors()

  def close(self):
    return self._impl.close()

  def __contextmanager__(
      self,
  ) -> Iterable[Self]:
    try:
      yield self
    finally:
      self.close()
